#!/usr/bin/env python

# Copyright 2016 99cloud Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from io import StringIO
from oslotest import base

from kayobe.plugins.action import merge_configs


TESTA = '''[DEFAULT]
key1 = b
       c
key2 = v1
       v2
key3 = v3
key3 = v4
key4 = v5

[b]
b_key1 = 1
b_key2 = 1
         2

[c]
c_key1 =
c_key2 = 1 2 3
         4 5 6

'''

TESTB = '''[DEFAULT]
key2 = v3
       v4
       v5
key4 = v4
key4 =

[b]
b_key2 = 2

'''

# TESTC is TESTA + TESTB
TESTC = '''[DEFAULT]
key1 = b
       c
key2 = v3
       v4
       v5
key3 = v3
key3 = v4
key4 = v4
key4 =

[b]
b_key1 = 1
b_key2 = 2

[c]
c_key1 =
c_key2 = 1 2 3
         4 5 6

'''

TESTA_NO_SECTIONS = '''key1 = a
key2 = b

'''

TESTB_NO_SECTIONS = '''key3 = c

'''

# TESTA_NO_SECTIONS and TESTB_NO_SECTIONS combined
TESTC_NO_SECTIONS = '''key1 = a
key2 = b
key3 = c

'''

TESTA_NO_DEFAULT_SECTION = '''key1 = a
key2 = b

[a]
key1 = not_a

[b]
key3 = not_c

'''

TESTB_NO_DEFAULT_SECTION = '''key3 = c

[b]
key2 = not_b
key3 = override

'''

# TESTA_NO_DEFAULT_SECTION and TESTB_NO_DEFAULT_SECTION combined
TESTC_NO_DEFAULT_SECTION = '''key1 = a
key2 = b
key3 = c

[a]
key1 = not_a

[b]
key3 = override
key2 = not_b

'''

# TESTC_NO_WHITESPACE is TESTA + TESTB without whitespace around equal signs
TESTC_NO_WHITESPACE = '''[DEFAULT]
key1=b
     c
key2=v3
     v4
     v5
key3=v3
key3=v4
key4=v4
key4=

[b]
b_key1=1
b_key2=2

[c]
c_key1=
c_key2=1 2 3
       4 5 6

'''


class OverrideConfigParserTest(base.BaseTestCase):

    def test_read_write(self):
        for ini in [TESTA,
                    TESTB,
                    TESTC,
                    TESTA_NO_SECTIONS,
                    TESTB_NO_SECTIONS,
                    TESTC_NO_SECTIONS,
                    TESTA_NO_DEFAULT_SECTION,
                    TESTB_NO_DEFAULT_SECTION,
                    TESTC_NO_DEFAULT_SECTION]:
            parser = merge_configs.OverrideConfigParser()
            parser.parse(StringIO(ini))
            output = StringIO()
            parser.write(output)
            self.assertEqual(ini, output.getvalue())
            output.close()

    def test_merge(self):
        parser = merge_configs.OverrideConfigParser()
        parser.parse(StringIO(TESTA))
        parser.parse(StringIO(TESTB))
        output = StringIO()
        parser.write(output)
        self.assertEqual(TESTC, output.getvalue())
        output.close()

    def test_merge_no_sections(self):
        parser = merge_configs.OverrideConfigParser()
        parser.parse(StringIO(TESTA_NO_SECTIONS))
        parser.parse(StringIO(TESTB_NO_SECTIONS))
        output = StringIO()
        parser.write(output)
        self.assertEqual(TESTC_NO_SECTIONS, output.getvalue())
        output.close()

    def test_merge_no_default_section(self):
        parser = merge_configs.OverrideConfigParser()
        parser.parse(StringIO(TESTA_NO_DEFAULT_SECTION))
        parser.parse(StringIO(TESTB_NO_DEFAULT_SECTION))
        output = StringIO()
        parser.write(output)
        self.assertEqual(TESTC_NO_DEFAULT_SECTION, output.getvalue())
        output.close()

    def test_merge_no_whitespace(self):
        parser = merge_configs.OverrideConfigParser(whitespace=False)
        parser.parse(StringIO(TESTA))
        parser.parse(StringIO(TESTB))
        output = StringIO()
        parser.write(output)
        self.assertEqual(TESTC_NO_WHITESPACE, output.getvalue())
        output.close()



# Copyright (c) 2021 StackHPC Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
This module provides Ansible filters that generate configuration for
systemd-networkd NetDevs, links and networks. The results are compatible with
the stackhpc.linux.systemd_networkd role.

Systemd-networkd uses INI-style configuration files, with the provision for
multiple sections with the same name, and multiple options with the same name
in a given section. This results in a slightly unwieldy data format used by the
role. The top level is a list of dicts with section names as keys. The values
are lists of dicts mapping option names to values.

Example schema (YAML):
- section1:
    - option1: value1
    - option2: value2
- section2
    - option3: value3
"""

import ipaddress

from ansible import errors
import jinja2

from kayobe.plugins.filter import networks
from kayobe.plugins.filter import utils


def _filter_options(config):
    """Filter out None values from a networkd config.

    :param config: List of sections to filter.
    :returns: a filtered list of sections without empty options.
    """
    # Example schema (YAML):
    # - section1:
    #     - option1: value1
    #     - option2:
    # - section2
    #     - option3:
    # We can filter this down to the following:
    # - section1:
    #     - option1: value1
    new_config = []
    for section_dict in config:
        new_section_dict = {}
        for section_name, section in section_dict.items():
            new_section = []
            for option_dict in section:
                new_option_dict = {}
                for option_name, option in option_dict.items():
                    if option is not None:
                        new_option_dict[option_name] = option
                if new_option_dict:
                    new_section.append(new_option_dict)
            if new_section:
                new_section_dict[section_name] = new_section
        if new_section_dict:
            new_config.append(new_section_dict)
    return new_config


def _ms_to_s(n):
    """Convert from milliseconds to seconds."""
    if n is not None:
        n = float(n) / 1000
    return n


def _vlan_netdev(context, name, inventory_hostname):
    """Return a networkd NetDev configuration for a VLAN interface.

    :param context: a Jinja2 Context object.
    :param name: name of the network.
    :param inventory_hostname: Ansible inventory hostname.
    """
    device = networks.net_interface(context, name, inventory_hostname)
    mtu = networks.net_mtu(context, name, inventory_hostname)
    vlan = networks.net_vlan(context, name, inventory_hostname)
    config = [
        {
            'NetDev': [
                {'Name': device},
                {'Kind': 'vlan'},
                {'MTUBytes': mtu},
            ],
        },
        {
            'VLAN': [
                {'Id': vlan},
            ]
        }
    ]
    return _filter_options(config)


def _bridge_netdev(context, name, inventory_hostname):
    """Return a networkd NetDev configuration for a bridge.

    :param context: a Jinja2 Context object.
    :param name: name of the network.
    :param inventory_hostname: Ansible inventory hostname.
    """
    device = networks.net_interface(context, name, inventory_hostname)
    mtu = networks.net_mtu(context, name, inventory_hostname)
    stp = networks.net_bridge_stp(context, name, inventory_hostname)
    config = [
        {
            'NetDev': [
                {'Name': device},
                {'Kind': 'bridge'},
                {'MTUBytes': mtu},
            ]
        }
    ]
    if stp is not None:
        config[0]['Bridge'] = [{'STP': stp}]
    return _filter_options(config)


def _bond_netdev(context, name, inventory_hostname):
    """Return a networkd NetDev configuration for a bond.

    :param context: a Jinja2 Context object.
    :param name: name of the network.
    :param inventory_hostname: Ansible inventory hostname.
    """
    device = networks.net_interface(context, name, inventory_hostname)
    mtu = networks.net_mtu(context, name, inventory_hostname)
    mode = networks.net_bond_mode(context, name, inventory_hostname)
    ad_select = networks.net_bond_ad_select(context, name, inventory_hostname)
    miimon = networks.net_bond_miimon(context, name, inventory_hostname)
    updelay = networks.net_bond_updelay(context, name, inventory_hostname)
    downdelay = networks.net_bond_downdelay(context, name, inventory_hostname)
    xmit_hash_policy = networks.net_bond_xmit_hash_policy(context, name,
                                                          inventory_hostname)
    lacp_rate = networks.net_bond_lacp_rate(context, name, inventory_hostname)
    config = [
        {
            'NetDev': [
                {'Name': device},
                {'Kind': 'bond'},
                {'MTUBytes': mtu},
            ]
        },
        {
            'Bond': [
                {'Mode': mode},
                {'AdSelect': ad_select},
                {'TransmitHashPolicy': xmit_hash_policy},
                {'LACPTransmitRate': lacp_rate},
                {'MIIMonitorSec': _ms_to_s(miimon)},
                {'UpDelaySec': _ms_to_s(updelay)},
                {'DownDelaySec': _ms_to_s(downdelay)},
            ]
        }
    ]
    return _filter_options(config)


def _veth_netdev(context, veth, inventory_hostname):
    """Return a networkd NetDev configuration for a veth pair.

    :param context: a Jinja2 Context object.
    :param veth: a dict describing the virtual Ethernet pair.
    :param inventory_hostname: Ansible inventory hostname.
    """
    interface = veth['name']
    peer = veth['peer']
    config = [
        {
            'NetDev': [
                {'Name': interface},
                {'Kind': 'veth'},
            ],
        },
        {
            'Peer': [
                {'Name': peer},
            ]
        }
    ]
    return _filter_options(config)


def _network_routes(routes, route_tables):
    """Return a list of routes for a networkd network.

    :param routes: a list of route dictionaries.
    :param route_tables: a dict mapping route table names to IDs.
    :returns: a list of routes for a networkd network.
    """
    return [
        {
            'Route': [
                # FIXME(mgoddard): No support for 'options'.
                {'Destination': route['cidr']},
                {'Gateway': route.get('gateway')},
                {'Table': route_tables.get(route.get('table'),
                                           route.get('table'))},
            ]
        }
        for route in routes or []
    ]


def _network_rules(rules, route_tables):
    """Return a list of routing policy rules for a networkd network.

    :param rules: a list of rule dictionaries.
    :param route_tables: a dict mapping route table names to IDs.
    :returns: a list of rules for a networkd network.
    """
    for rule in rules or []:
        if not isinstance(rule, dict):
            raise errors.AnsibleFilterError(
                "Routing policy rules must be defined in dictionary "
                "format for systemd-networkd")
    return [
        {
            'RoutingPolicyRule': [
                {'From': rule.get("from")},
                {'To': rule.get("to")},
                {'Priority': rule.get("priority")},
                {'Table': route_tables.get(rule.get('table'),
                                           rule.get('table'))},
            ]
        }
        for rule in rules or []
    ]


def _network(context, name, inventory_hostname, bridge, bond, vlan_interfaces):
    """Return a networkd network for an interface.

    :param context: a Jinja2 Context object.
    :param name: name of the network.
    :param inventory_hostname: Ansible inventory hostname.
    :param bridge: Name of a bridge into which the interface is plugged, or
                   None.
    :param bond: Name of a bond of which the interface is a member, or None.
    :param vlan_interfaces: List of VLAN subinterfaces of the interface.
    """
    # FIXME(mgoddard): Currently does not support: ethtool_opts, zone,
    # allowed_addresses.
    device = networks.net_interface(context, name, inventory_hostname)
    ip = networks.net_ip(context, name, inventory_hostname)
    cidr = networks.net_cidr(context, name, inventory_hostname)
    gateway = networks.net_gateway(context, name, inventory_hostname)
    if ip is None:
        gateway = None
    else:
        if not cidr:
            raise errors.AnsibleFilterError(
                "No CIDR attribute configured for '%s' network but it has an "
                "IP address" %
                (name))
        ip = "%s/%s" % (ip, ipaddress.ip_network(cidr).prefixlen)

    mtu = networks.net_mtu(context, name, inventory_hostname)
    routes = networks.net_routes(context, name, inventory_hostname)
    rules = networks.net_rules(context, name, inventory_hostname)
    bootproto = networks.net_bootproto(context, name, inventory_hostname)
    defroute = networks.net_defroute(context, name, inventory_hostname)
    if defroute is not None:
        defroute = utils.call_bool_filter(context, defroute)
    config = [
        {
            'Match': [
                {'Name': device},
            ]
        },
        {
            'Network': [
                {'Address': ip},
                {'Gateway': gateway},
                {'DHCP': ('yes' if bootproto and bootproto.lower() == 'dhcp'
                          else None)},
                {'UseGateway': ('false'
                                if defroute is not None and not defroute
                                else None)},
                {'Bridge': bridge},
                {'Bond': bond},
            ] + [
                {'VLAN': vlan_interface}
                for vlan_interface in vlan_interfaces
            ]
        },
        {
            'Link': [
                {'MTUBytes': mtu},
            ]
        },
    ]

    # NOTE(mgoddard): Systemd-networkd does not support named route tables
    # until v248. Until then, translate names to numeric IDs using the
    # network_route_tables variable.
    route_tables = utils.get_hostvar(context, "network_route_tables",
                                     inventory_hostname)
    route_tables = {table["name"]: table["id"] for table in route_tables}
    config += _network_routes(routes, route_tables)
    config += _network_rules(rules, route_tables)

    return _filter_options(config)


def _vlan_parent_network(device, mtu, vlan_interfaces):
    """Return a networkd network configuration for a VLAN parent interface.

    :param device: name of the interface.
    :param mtu: Interface MTU.
    :param vlan_interfaces: List of VLAN subinterfaces of the interface.
    """
    config = [
        {
            'Match': [
                {'Name': device},
            ]
        },
        {
            'Network': [
                {'VLAN': vlan_interface}
                for vlan_interface in vlan_interfaces
            ]
        },
        {
            'Link': [
                {'MTUBytes': mtu},
            ]
        }
    ]
    return _filter_options(config)


def _bridge_port_network(context, name, port, inventory_hostname,
                         vlan_interfaces):
    """Return a networkd network configuration for a bridge port.

    :param context: a Jinja2 Context object.
    :param name: name of the network.
    :param port: name of the bridge port interface.
    :param inventory_hostname: Ansible inventory hostname.
    :param vlan_interfaces: List of VLAN subinterfaces of the interface.
    """
    bridge = networks.get_and_validate_interface(context, name,
                                                 inventory_hostname)
    mtu = networks.net_mtu(context, name, inventory_hostname)
    config = [
        {
            'Match': [
                {'Name': port},
            ]
        },
        {
            'Network': [
                {'Bridge': bridge},
            ] + [
                {'VLAN': vlan_interface}
                for vlan_interface in vlan_interfaces
            ]
        },
        {
            'Link': [
                {'MTUBytes': mtu},
            ]
        }
    ]
    return _filter_options(config)


def _bond_member_network(context, name, member, inventory_hostname,
                         vlan_interfaces):
    """Return a networkd network configuration for a bond member.

    :param context: a Jinja2 Context object.
    :param name: name of the network.
    :param member: name of the bond member interface.
    :param inventory_hostname: Ansible inventory hostname.
    :param vlan_interfaces: List of VLAN subinterfaces of the interface.
    """
    bond = networks.get_and_validate_interface(context, name,
                                               inventory_hostname)
    mtu = networks.net_mtu(context, name, inventory_hostname)
    config = [
        {
            'Match': [
                {'Name': member},
            ]
        },
        {
            'Network': [
                {'Bond': bond},
            ] + [
                {'VLAN': vlan_interface}
                for vlan_interface in vlan_interfaces
            ]
        },
        {
            'Link': [
                {'MTUBytes': mtu},
            ]
        }
    ]
    return _filter_options(config)


def _veth_network(context, veth, inventory_hostname):
    """Return a networkd network configuration for a veth link.

    :param context: a Jinja2 Context object.
    :param veth: a dict describing the virtual Ethernet pair.
    :param inventory_hostname: Ansible inventory hostname.
    """
    interface = veth['name']
    bridge = veth['bridge']
    mtu = veth['mtu']
    config = [
        {
            'Match': [
                {'Name': interface},
            ]
        },
        {
            'Network': [
                {'Bridge': bridge},
            ]
        },
        {
            'Link': [
                {'MTUBytes': mtu},
            ]
        }
    ]
    return _filter_options(config)


def _veth_peer_network(context, veth, inventory_hostname):
    """Return a networkd network configuration for a veth peer.

    :param context: a Jinja2 Context object.
    :param veth: a dict describing the virtual Ethernet pair.
    :param inventory_hostname: Ansible inventory hostname.
    """
    interface = veth['peer']
    mtu = veth['mtu']
    config = [
        {
            'Match': [
                {'Name': interface},
            ]
        },
        {
            'Network': [
                # NOTE(mgoddard): bring the interface up, even without an IP.
                {'ConfigureWithoutCarrier': 'true'},
            ]
        },
        {
            'Link': [
                {'MTUBytes': mtu},
            ]
        }
    ]
    return _filter_options(config)


def _ether_link(context, name, inventory_hostname):
    """Return a networkd link configuration for a ether.

    :param context: a Jinja2 Context object.
    :param name: name of the network.
    :param inventory_hostname: Ansible inventory hostname.
    """
    config = []

    device = networks.net_interface(context, name, inventory_hostname)
    macaddress = networks.net_macaddress(context, name, inventory_hostname)

    if macaddress is not None:
        config = [
            {
                'Match': [
                    {'PermanentMACAddress': macaddress},
                ],
            },
            {
                'Link': [
                    {'Name': device},
                ],
            }
        ]

    return _filter_options(config)


def _add_to_result(result, prefix, device, config):
    """Add configuration for an interface to a filter result.

    :param result: the result dict.
    :param prefix: the systemd-networkd configuration file prefix.
    :param device: the interface being configured.
    :param config: the configuration to add to the result.
    :raises: AnsibleFilterError if the interface already exists in the result.
    """
    key = "%s%s" % (prefix, device)
    # Catch the case where interface configuration is added multiple times.
    # This should not happen.
    if key in result:
        raise errors.AnsibleFilterError(
            "Programming error: duplicate interface configuration for %s: "
            "have %s"
            % (device, result))
    result[key] = config


@jinja2.pass_context
def networkd_netdevs(context, names, inventory_hostname=None):
    """Return a dict representation of networkd NetDev configuration.

    The format is compatible with the systemd_networkd_netdev variable in the
    stackhpc.linux.systemd_networkd role.

    :param context: a Jinja2 Context object.
    :param names: List of names of networks.
    :param inventory_hostname: Ansible inventory hostname.
    :returns: a dict representation of networkd NetDev configuration.
    """
    # Prefix for configuration file names.
    prefix = utils.get_hostvar(context, "networkd_prefix", inventory_hostname)

    result = {}

    # VLANs.
    for name in networks.net_select_vlan_interfaces(context, names,
                                                    inventory_hostname):
        device = networks.get_and_validate_interface(context, name,
                                                     inventory_hostname)
        netdev = _vlan_netdev(context, name, inventory_hostname)
        _add_to_result(result, prefix, device, netdev)

    # Bridges.
    for name in networks.net_select_bridges(context, names,
                                            inventory_hostname):
        device = networks.get_and_validate_interface(context, name,
                                                     inventory_hostname)
        netdev = _bridge_netdev(context, name, inventory_hostname)
        _add_to_result(result, prefix, device, netdev)

    # Bonds.
    for name in networks.net_select_bonds(context, names, inventory_hostname):
        device = networks.get_and_validate_interface(context, name,
                                                     inventory_hostname)
        netdev = _bond_netdev(context, name, inventory_hostname)
        _add_to_result(result, prefix, device, netdev)

    # Virtual Ethernet pairs.
    veths = networks.get_ovs_veths(context, names, inventory_hostname)
    for veth in veths:
        netdev = _veth_netdev(context, veth, inventory_hostname)
        device = veth['name']
        _add_to_result(result, prefix, device, netdev)

    return result


@jinja2.pass_context
def networkd_links(context, names, inventory_hostname=None):
    """Return a dict representation of networkd link configuration.

    The format is compatible with the systemd_networkd_link variable in the
    stackhpc.linux.systemd_networkd role.

    :param context: a Jinja2 Context object.
    :param names: List of names of networks.
    :param inventory_hostname: Ansible inventory hostname.
    :returns: a dict representation of networkd link configuration.
    """
    # Prefix for configuration file names.
    prefix = utils.get_hostvar(context, "networkd_prefix", inventory_hostname)

    result = {}

    # only ethers
    for name in networks.net_select_ethers(context, names, inventory_hostname):
        device = networks.get_and_validate_interface(context, name,
                                                     inventory_hostname)
        ether_link = _ether_link(context, name, inventory_hostname)
        if ether_link:
            _add_to_result(result, prefix, device, ether_link)

    return result


@jinja2.pass_context
def networkd_networks(context, names, inventory_hostname=None):
    """Return a dict representation of networkd network configuration.

    The format is compatible with the systemd_networkd_network variable in the
    stackhpc.linux.systemd_networkd role.

    :param context: a Jinja2 Context object.
    :param names: List of names of networks.
    :param inventory_hostname: Ansible inventory hostname.
    :returns: a dict representation of networkd network configuration.
    """
    # TODO(mgoddard): some attributes are currently not supported for
    # systemd-networkd: rules, route options, ethtool_opts, zone,
    # allowed addresses

    # Build up some useful mappings.
    bridge_port_to_bridge = {}
    bond_member_to_bond = {}
    interface_to_vlans = {}

    # List of all interfaces.
    interfaces = [
        networks.net_interface(context, name, inventory_hostname)
        for name in names
    ]

    # Map bridge ports to bridges.
    for name in networks.net_select_bridges(context, names,
                                            inventory_hostname):
        device = networks.get_and_validate_interface(context, name,
                                                     inventory_hostname)
        for port in networks.net_bridge_ports(context, name,
                                              inventory_hostname):
            bridge_port_to_bridge[port] = device

    # Map bond members to bonds.
    for name in networks.net_select_bonds(context, names, inventory_hostname):
        device = networks.get_and_validate_interface(context, name,
                                                     inventory_hostname)
        for member in networks.net_bond_slaves(context, name,
                                               inventory_hostname):
            bond_member_to_bond[member] = device

    # Map interfaces to lists of VLAN subinterfaces.
    for name in networks.net_select_vlans(context, names, inventory_hostname):
        device = networks.get_and_validate_interface(context, name,
                                                     inventory_hostname)
        vlan = networks.net_vlan(context, name, inventory_hostname)
        mtu = networks.net_mtu(context, name, inventory_hostname)
        parent = networks.get_vlan_parent(
            context, name, device, vlan, inventory_hostname)
        vlan_interfaces = interface_to_vlans.setdefault(parent, [])
        vlan_interfaces.append({"device": device, "mtu": mtu})

    # Prefix for configuration file names.
    prefix = utils.get_hostvar(context, "networkd_prefix", inventory_hostname)

    result = {}

    # Configured networks.
    for name in names:
        device = networks.get_and_validate_interface(context, name,
                                                     inventory_hostname)
        bridge = bridge_port_to_bridge.get(device)
        bond = bond_member_to_bond.get(device)
        vlan_interfaces = interface_to_vlans.get(device, [])
        net = _network(context, name, inventory_hostname, bridge, bond,
                       [vlan["device"] for vlan in vlan_interfaces])
        _add_to_result(result, prefix, device, net)

    # VLAN parent interfaces that are not in configured networks, bridge ports
    # or bond members.
    implied_vlan_parents = (set(interface_to_vlans) -
                            set(interfaces) -
                            set(bridge_port_to_bridge) -
                            set(bond_member_to_bond))
    for device in implied_vlan_parents:
        vlan_interfaces = interface_to_vlans[device]
        vlan_mtus = [vlan["mtu"] for vlan in vlan_interfaces if vlan["mtu"]]
        mtu = max(vlan_mtus) if vlan_mtus else None
        net = _vlan_parent_network(device, mtu,
                                   [vlan["device"]
                                    for vlan in vlan_interfaces])
        _add_to_result(result, prefix, device, net)

    # Bridge ports that are not in configured networks.
    for name in networks.net_select_bridges(context, names,
                                            inventory_hostname):
        device = networks.get_and_validate_interface(context, name,
                                                     inventory_hostname)
        bridge_ports = networks.net_bridge_ports(context, name,
                                                 inventory_hostname)
        for port in set(bridge_ports) - set(interfaces):
            vlan_interfaces = interface_to_vlans.get(port, [])
            net = _bridge_port_network(context, name, port, inventory_hostname,
                                       [vlan["device"]
                                        for vlan in vlan_interfaces])
            _add_to_result(result, prefix, port, net)

    # Bond members that are not in configured networks.
    for name in networks.net_select_bonds(context, names, inventory_hostname):
        device = networks.get_and_validate_interface(context, name,
                                                     inventory_hostname)
        bond_members = networks.net_bond_slaves(context, name,
                                                inventory_hostname)
        for member in set(bond_members) - set(interfaces):
            vlan_interfaces = interface_to_vlans.get(member, [])
            net = _bond_member_network(context, name, member,
                                       inventory_hostname,
                                       [vlan["device"]
                                        for vlan in vlan_interfaces])
            _add_to_result(result, prefix, member, net)

    # Virtual Ethernet pairs for Open vSwitch.
    veths = networks.get_ovs_veths(context, names, inventory_hostname)
    for veth in veths:
        net = _veth_network(context, veth, inventory_hostname)
        device = veth['name']
        _add_to_result(result, prefix, device, net)

        net = _veth_peer_network(context, veth, inventory_hostname)
        device = veth['peer']
        _add_to_result(result, prefix, device, net)

    return result


def get_filters():
    return {
        'networkd_netdevs': networkd_netdevs,
        'networkd_links': networkd_links,
        'networkd_networks': networkd_networks,
    }



# Copyright (c) 2018 StackHPC Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import os.path

from kayobe.tests.molecule import utils

import pytest
import testinfra.utils.ansible_runner
import yaml


testinfra_hosts = testinfra.utils.ansible_runner.AnsibleRunner(
    os.environ['MOLECULE_INVENTORY_FILE']).get_hosts('all')

@pytest.mark.parametrize(
    'path',
    ['aodh.conf',
     'barbican.conf',
     'cinder.conf',
     'cloudkitty.conf',
     'designate.conf',
     'galera.cnf',
     'glance.conf',
     'grafana.ini',
     'heat.conf',
     'ironic.conf',
     'ironic-inspector.conf',
     'keystone.conf',
     'magnum.conf',
     'manila.conf',
     'masakari.conf',
     'multipath.conf',
     'neutron/ml2_conf.ini',
     'neutron.conf',
     'nova.conf',
     'octavia.conf',
     'placement.conf',
     'backup.my.cnf'])
def test_service_ini_file(host, path):
    # TODO(mgoddard): Check more of config file contents.
    # Tests config added with extra vars e.g kolla_extra_aodh. I.e the
    # the internal role templates.
    path = os.path.join('/etc/kolla/config', path)
    extra_section = 'extra-%s' % os.path.basename(path)
    expected = {extra_section: {'foo': 'bar'}}
    utils.test_ini_file(host, path, expected=expected)

@pytest.mark.parametrize(
    'path',
    ['aodh.conf',
     'barbican.conf',
     'cinder.conf',
     'cloudkitty.conf',
     'designate.conf',
     'galera.cnf',
     'glance.conf',
     'grafana.ini',
     'heat.conf',
     'ironic.conf',
     'ironic-inspector.conf',
     'keystone.conf',
     'magnum.conf',
     'manila.conf',
     'masakari.conf',
     'neutron/ml2_conf.ini',
     'neutron.conf',
     'nova.conf',
     'octavia.conf',
     'placement.conf',
     'backup.my.cnf'])
def test_service_ini_file_extra_confs(host, path):
    # Tests config added via extra config files
    path = os.path.join('/etc/kolla/config', path)
    extra_section = 'extra-file-%s' % os.path.basename(path)
    expected = {extra_section: {'bar': 'baz'}}
    utils.test_ini_file(host, path, expected=expected)

@pytest.mark.parametrize(
    'path',
    ['ironic/ironic-agent.initramfs',
     'ironic/ironic-agent.kernel',
     'nova/nova-libvirt/cacert.pem',
     'nova/nova-libvirt/clientcert.pem',
     'nova/nova-libvirt/clientkey.pem'])
def test_service_non_ini_file(host, path):
    # TODO(mgoddard): Check config file contents.
    path = os.path.join('/etc/kolla/config', path)
    utils.test_file(host, path)

@pytest.mark.parametrize(
    'path,regex',
    [('fluentd/input/01-test.conf', 'grepme')])
def test_service_non_ini_file_regex(host, path, regex):
    path = os.path.join('/etc/kolla/config', path)
    utils.test_regex_in_file(host, path, regex=regex)

@pytest.mark.parametrize(
    'relative_path',
    ['aodh/dummy.yml',
     'opensearch.yml',
     'prometheus/prometheus.yml.d/dummy.yml'])
def test_service_extra_yml_config(host, relative_path):
    path = os.path.join('/etc/kolla/config', relative_path)
    utils.test_file(host, path)
    content = yaml.safe_load(host.file(path).content_string)
    assert content["dummy_variable"] == 123

def test_service_extra_ini_config(host):
    relative_path = "aodh/dummy.ini"
    path = os.path.join('/etc/kolla/config', relative_path)
    utils.test_file(host, path)
    expected = {
        "dummy-section": {"dummy_variable": "123"}
    }
    utils.test_ini_file(host, path, expected=expected)



# Copyright (c) 2017 StackHPC Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import base64
from collections import defaultdict
import configparser
import glob
import graphlib
from importlib.metadata import Distribution
import json
import logging
import os
import shutil
import subprocess
import sys

from ansible.parsing.yaml.loader import AnsibleLoader
import yaml

from kayobe import exception


LOG = logging.getLogger(__name__)


def get_data_files_path(*relative_path):
    """Given a relative path to a data file, return the absolute path"""
    # Detect editable pip install / python setup.py develop and use a path
    # relative to the source directory
    return os.path.join(_get_base_path(), *relative_path)


def _detect_install_prefix(path):
    script_path = os.path.realpath(path)
    script_path = os.path.normpath(script_path)
    components = script_path.split(os.sep)
    # use heuristic: anything before the last 'lib' in path is the prefix
    if 'lib' not in components:
        return None
    last_lib = len(components) - 1 - components[::-1].index('lib')
    prefix = components[:last_lib]
    prefix_path = os.sep.join(prefix)
    return prefix_path


def _get_direct_url_if_editable(dist):
    direct_url = os.path.join(dist._path, 'direct_url.json')
    editable = None
    if os.path.isfile(direct_url):
        with open(direct_url, 'r') as f:
            direct_url_content = json.loads(f.readline().strip())
            dir_info = direct_url_content.get('dir_info')
            if dir_info is not None:
                editable = dir_info.get('editable')
            if editable:
                url = direct_url_content['url']
                prefix = 'file://'
                if url.startswith(prefix):
                    return url[len(prefix):]

    return None


def _get_base_path():
    override = os.environ.get("KAYOBE_DATA_FILES_PATH")
    if override:
        return os.path.join(override)

    kayobe_dist = list(Distribution.discover(name="kayobe"))
    if kayobe_dist:
        direct_url = _get_direct_url_if_editable(kayobe_dist[0])
        if direct_url:
            return direct_url

    egg_glob = os.path.join(
        sys.prefix, 'lib*', 'python*', '*-packages', 'kayobe.egg-link'
    )
    egg_link = glob.glob(egg_glob)
    if egg_link:
        with open(egg_link[0], "r") as f:
            realpath = f.readline().strip()
        return os.path.join(realpath)

    prefix = _detect_install_prefix(__file__)
    if prefix:
        return os.path.join(prefix, "share", "kayobe")

    # Assume uninstalled
    return os.path.join(os.path.dirname(os.path.realpath(__file__)), "..")


def galaxy_role_install(role_file, roles_path, force=False):
    """Install Ansible roles via Ansible Galaxy."""
    cmd = ["ansible-galaxy", "role", "install"]
    cmd += ["--roles-path", roles_path]
    cmd += ["--role-file", role_file]
    if force:
        cmd += ["--force"]
    try:
        run_command(cmd)
    except subprocess.CalledProcessError as e:
        LOG.error("Failed to install Ansible roles from %s via Ansible "
                  "Galaxy: returncode %d", role_file, e.returncode)
        sys.exit(e.returncode)


def galaxy_collection_install(requirements_file, collections_path,
                              force=False):
    requirements = read_yaml_file(requirements_file)
    if not isinstance(requirements, dict):
        # Handle legacy role list format, which causes the command to fail.
        return
    cmd = ["ansible-galaxy", "collection", "install"]
    cmd += ["--collections-path", collections_path]
    cmd += ["--requirements-file", requirements_file]
    env_defaults = {
        # NOTE(wszumski): Allow overriding of ansible builtin collections in
        # kayobe requirements.yml.
        "ANSIBLE_COLLECTIONS_SCAN_SYS_PATH": "False",
    }
    env = env_defaults | os.environ
    if force:
        cmd += ["--force"]
    try:
        run_command(cmd, env=env)
    except subprocess.CalledProcessError as e:
        LOG.error("Failed to install Ansible collections from %s via Ansible "
                  "Galaxy: returncode %d", requirements_file, e.returncode)
        sys.exit(e.returncode)


def galaxy_remove(roles_to_remove, roles_path):

    """Remove Ansible roles via Ansible Galaxy."""
    cmd = ["ansible-galaxy", "role", "remove"]
    cmd += ["--roles-path", roles_path]
    cmd += roles_to_remove
    try:
        run_command(cmd)
    except subprocess.CalledProcessError as e:
        LOG.error("Failed to remove Ansible roles %s via Ansible "
                  "Galaxy: returncode %d",
                  ",".join(roles_to_remove), e.returncode)
        sys.exit(e.returncode)


def read_file(path, mode="r"):
    """Read the content of a file."""
    with open(path, mode) as f:
        return f.read()


def read_yaml_file(path):
    """Read and decode a YAML file."""
    try:
        content = read_file(path)
    except IOError as e:
        print("Failed to open YAML file %s: %s" %
              (path, repr(e)))
        sys.exit(1)
    try:
        return yaml.safe_load(content)
    except yaml.YAMLError as e:
        print("Failed to decode YAML file %s: %s" %
              (path, repr(e)))
        sys.exit(1)


def read_config_dump_yaml_file(path):
    """Read and decode a configuration dump YAML file."""
    try:
        content = read_file(path)
    except IOError as e:
        print("Failed to open config dump file %s: %s" %
              (path, repr(e)))
        sys.exit(1)
    try:
        # AnsibleLoader supports loading vault encrypted variables.
        return AnsibleLoader(content).get_single_data()
    except yaml.YAMLError as e:
        print("Failed to decode config dump YAML file %s: %s" %
              (path, repr(e)))
        sys.exit(1)


def is_readable_dir(path):
    """Check whether a path references a readable directory."""
    if not os.path.exists(path):
        return {"result": False, "message": "Path does not exist"}
    if not os.path.isdir(path):
        return {"result": False, "message": "Path is not a directory"}
    if not os.access(path, os.R_OK):
        return {"result": False, "message": "Directory is not readable"}
    return {"result": True}


def is_readable_file(path):
    """Check whether a path references a readable file."""
    if not os.path.exists(path):
        return {"result": False, "message": "Path does not exist"}
    if not os.path.isfile(path):
        return {"result": False, "message": "Path is not a file"}
    if not os.access(path, os.R_OK):
        return {"result": False, "message": "File is not readable"}
    return {"result": True}


def run_command(cmd, quiet=False, check_output=False, **kwargs):
    """Run a command, checking the output.

    :param quiet: Redirect output to /dev/null
    :param check_output: Whether to return the output of the command
    :returns: The output of the command if check_output is true
    """
    if isinstance(cmd, str):
        cmd_string = cmd
    else:
        cmd_string = " ".join(cmd)
    LOG.debug("Running command: %s", cmd_string)
    if quiet:
        kwargs["stderr"] = subprocess.DEVNULL
        if not check_output:
            kwargs["stdout"] = subprocess.DEVNULL
    if check_output:
        return subprocess.check_output(cmd, **kwargs)
    else:
        subprocess.check_call(cmd, **kwargs)


def quote_and_escape(value):
    """Quote and escape a string.

    Adds enclosing single quotes to the string passed, and escapes single
    quotes within the string using backslashes. This is useful for passing
    'extra vars' to Ansible. Without this, Ansible only uses the part of the
    string up to the first whitespace.

    :param value: the string to quote and escape.
    :returns: the quoted and escaped string.
    """
    if not isinstance(value, str):
        return value
    return "'" + value.replace("'", "'\\''") + "'"


def escape_jinja(string):
    """Escapes a string so that jinja template variables are not expanded

    :param string: the string to escape
    :return: the escaped string
    """
    # We base64 encode the string to avoid the need to escape characters.
    # This is because ansible has some parsing quirks that makes it fairly
    # hard to escape stuff in generic way.
    # See: https://github.com/ansible/ansible/issues/10464

    b64_value = base64.b64encode(string.encode())
    return ''.join(('{{', "'", b64_value.decode(), "' | b64decode ", '}}'))


def intersect_limits(args_limit, cli_limit):
    """Create an Ansible host pattern of the intersection of two patterns.

    :param args_limit: user-specified limit, or None.
    :param cli_limit: limit originating from this CLI, or None.
    :returns: a string representing an intersection of the two patterns.
    """
    # NOTE(mgoddard): Ansible uses either commas (,) or colons (:) to separate
    # parts of a host pattern. An intersection is specified using a separator
    # followed by an ampersand (&). If a mix of comma and colon separators is
    # used, Ansible picks one and treats the other as part of the host pattern.
    # This leads to hard to diagnose errors. Try to determine which separator
    # the user has specified, and be consistent. Error if both are used.
    if args_limit and ',' in args_limit:
        if ':' in args_limit:
            raise exception.Error("Invalid format for host limit argument. "
                                  "Cannot mix commas and colons to separate "
                                  "hosts")
        separator = ',&'
    else:
        separator = ':&'
    limits = [limit for limit in [args_limit, cli_limit] if limit]
    return separator.join(limits)


def copy_dir(src, dest, exclude=None):
    """Copy recursively a directory.

    :param src: path of the source directory
    :param dest: destination path, will be created if it does not exist
    :param exclude: names of files or directories at the root of the source
                    directory to exclude during copy
    """
    if exclude is None:
        exclude = []

    if not os.path.isdir(dest):
        os.mkdir(dest)

    for file in os.listdir(src):
        if file in exclude:
            continue

        src_path = os.path.join(src, file)
        dest_path = os.path.join(dest, file)
        if os.path.isdir(src_path):
            copy_dir(src_path, dest_path)
        else:
            shutil.copy2(src_path, dest_path)


def get_kayobe_environment_path(base_path, environment):
    """Return the path to the Kayobe environment or None if not specified."""
    env_path = None
    if environment:
        # Specified via --environment or KAYOBE_ENVIRONMENT.
        kc_environments = os.path.join(base_path, "environments")
        env_path = os.path.join(kc_environments, environment)
    return env_path


class EnvironmentFinder(object):
    """Dependency resolver for kayobe environments

    The constraints are specified via a .kayobe-environment file.
    """

    def __new__(cls, base_path, environment):
        # Singleton instance so we don't have to resolve dependencies multiple
        # times or pass round a single instance.
        it = cls.__dict__.get("__it__")
        if it is None:
            it = {}
        if (base_path, environment) in it:
            return it[(base_path, environment)]
        singleton = object.__new__(cls)
        singleton._init(base_path, environment)
        it[(base_path, environment)] = singleton
        return singleton

    def _init(self, base_path, environment):
        self._base_path = base_path
        self._environment = environment
        self._ordering = None

    @staticmethod
    def _read_metadata(path):
        if os.path.exists(path) and os.path.isfile(path):
            metadata = read_yaml_file(path)
            return metadata
        return {}

    def _collect(self, environment, result, visited):
        # Updates result to contain dependency graph
        base = self._base_path
        env_path = os.path.join(base, 'environments', environment)
        dot_environment_path = os.path.join(env_path, '.kayobe-environment')
        if dot_environment_path in visited:
            return
        visited.add(dot_environment_path)
        metadata = EnvironmentFinder._read_metadata(dot_environment_path)
        dependencies = metadata.get("dependencies", [])
        if not isinstance(dependencies, list):
            raise exception.Error(".kayobe-environment: dependencies field "
                                  "should be a list")
        result[environment] |= set(dependencies)
        for dependency in dependencies:
            if not isinstance(dependency, str):
                raise exception.Error("Kayobe environment dependency items "
                                      "should be strings")
            self._collect(dependency, result, visited)

    def ordered(self):
        """List of environments ordered by the constraints"""
        environment = self._environment
        if not environment:
            return []
        if self._ordering is not None:
            return self._ordering.copy()
        graph = defaultdict(set)
        self._collect(environment, graph, set())
        ts = graphlib.TopologicalSorter(graph)
        try:
            ordering = list(ts.static_order())
        except graphlib.CycleError as e:
            # https://docs.python.org/3/library/graphlib.html#graphlib.CycleError
            cycle = e.args[1]
            raise exception.Error("You have created a cycle with your "
                                  "environment dependencies. Please break "
                                  "this cycle and try again. The cycle is: %s"
                                  % cycle)
        self._ordering = ordering if ordering else [environment]
        return self._ordering.copy()

    def ordered_paths(self):
        """Paths to each environment ordered by the constraints"""
        result = []
        environments = self.ordered()
        for environment in environments:
            full_path = get_kayobe_environment_path(
                self._base_path,
                environment
            )
            result.append(full_path)
        return result


def _gitreview_is_kayobe_config(gitreview_path):
    """Return whether a .gitreview file is for kayobe-config."""
    config = configparser.ConfigParser()
    config_string = read_file(gitreview_path)
    config.read_string(config_string)
    gerrit_project = config.get('gerrit', 'project')
    if not gerrit_project:
        return False
    gerrit_project = os.path.basename(gerrit_project)
    gerrit_project = os.path.splitext(gerrit_project)[0]
    if gerrit_project == 'kayobe-config':
        return True


def validate_config_path(config_path):
    """Validate the Kayobe configuration path.

    Check whether we are executing from inside a Kayobe configuration
    repository, and if so, assert that matches the Kayobe configuration path
    defined in CLI args or environment variables.

    Exit 1 if validation fails.

    :param config_path: Kayobe configuration path or None.
    """
    assert config_path

    try:
        cmd = ["git", "rev-parse", "--show-toplevel"]
        repo_root = run_command(cmd, quiet=True, check_output=True)
    except (FileNotFoundError, subprocess.CalledProcessError):
        # FileNotFoundError: git probably not installed.
        # CalledProcessError: probably not in a git repository.
        return

    repo_root = repo_root.decode().strip()
    if config_path:
        repo_config_path = os.path.join(repo_root, "etc", "kayobe")
        if repo_config_path == os.path.realpath(config_path):
            return

    # Paths did not match. Check that repo_root does not look like a Kayobe
    # configuration repo.
    gitreview_path = os.path.join(repo_root, ".gitreview")
    result = is_readable_file(gitreview_path)
    if result["result"]:
        if _gitreview_is_kayobe_config(gitreview_path):
            LOG.error("Executing from within a different Kayobe configuration "
                      "repository is not allowed")
            sys.exit(1)



# Copyright (c) 2023 StackHPC Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from ansible.errors import AnsibleError
from ansible.plugins.loader import lookup_loader
from ansible.plugins.lookup import LookupBase

from kayobe.utils import EnvironmentFinder

__version__ = "1.0.0"

try:
    from __main__ import display
except ImportError:
    from ansible.utils.display import Display

    display = Display()


class LookupModule(LookupBase):
    def run(self, terms, variables=None, **kwargs):
        lookup = lookup_loader.get(
            'vars', loader=self._loader, templar=self._templar
        )
        # Values in variables are untemplated, e.g:
        # {{ lookup('env', 'KAYOBE_CONFIG_PATH') | default('/etc/kayobe', true) }}  # noqa
        environment = lookup.run(
            ["kayobe_environment"],
            variables=variables, default='')[0]
        kayobe_config_path = lookup.run(
            ["kayobe_config_path"],
            variables=variables, default='')[0]
        if not environment:
            return []
        if not kayobe_config_path:
            raise AnsibleError("kayobe_config_path is unset")
        environment_finder = EnvironmentFinder(kayobe_config_path, environment)
        return environment_finder.ordered_paths()



# Copyright (c) 2017 StackHPC Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

DOCUMENTATION = """
module: ip_allocation
short_description: Allocate an IP address for a host from a pool
author: Mark Goddard (mark@stackhpc.com)
options:
  - option-name: net_name
    description: Name of the network
    required: True
    type: string
  - option-name: hostname
    description: Name of the host
    required: True
    type: string
  - option-name: cidr
    description: IP Network in CIDR format
    required: True
    type: string
  - option-name: allocation_pool_start
    description: First address of the pool from which to allocate
    required: False
    type: string
  - option-name: allocation_pool_end
    description: Last address of the pool from which to allocate
    required: False
    type: string
  - option-name: allocation_file
    description: >
      Path to a file in which to store the allocations. Will be created if it
      does not exist.
    required: True
    type: string
requirements:
  - netaddr
  - PyYAML
"""

EXAMPLES = """
- name: Ensure host has an IP address
  ip_allocation:
    net_name: my-network
    hostname: my-host
    cidr: 10.0.0.0/24
    allocation_pool_start: 10.0.0.1
    allocation_pool_end: 10.0.0.254
    allocation_file: /path/to/allocation/file.yml
"""

RETURN = """
ip:
  description: The allocated IP address
  returned: success
  type: string
  sample: 10.0.0.1
"""

from ansible.module_utils.basic import *
import sys

# Store a list of import errors to report to the user.
IMPORT_ERRORS=[]
try:
    import netaddr
except Exception as e:
    IMPORT_ERRORS.append(e)
try:
    import yaml
except Exception as e:
    IMPORT_ERRORS.append(e)


def read_allocations(module):
    """Read IP address allocations from the allocation file."""
    filename = module.params['allocation_file']
    try:
        with open(filename, 'r') as f:
            content = yaml.safe_load(f)
    except IOError as e:
        if e.errno == errno.ENOENT:
            # Ignore ENOENT - we will create the file.
            return {}
        module.fail_json(msg="Failed to open allocation file %s for reading" % filename)
    except yaml.YAMLError as e:
        module.fail_json(msg="Failed to parse allocation file %s as YAML" % filename)
    if content is None:
        # If the file is empty, yaml.safe_load() will return None.
        content = {}
    return content


def write_allocations(module, allocations):
    """Write IP address allocations to the allocation file."""
    filename = module.params['allocation_file']
    try:
        with open(filename, 'w') as f:
            yaml.dump(allocations, f, default_flow_style=False)
    except IOError as e:
        module.fail_json(msg="Failed to open allocation file %s for writing" % filename)
    except yaml.YAMLError as e:
        module.fail_json(msg="Failed to dump allocation file %s as YAML" % filename)


def update_allocation(module, allocations):
    """Allocate an IP address on a network for a host.

    :param module: AnsibleModule instance
    :param allocations: Existing IP address allocations
    """
    net_name = module.params['net_name']
    hostname = module.params['hostname']
    cidr = module.params['cidr']
    allocation_pool_start = module.params['allocation_pool_start']
    allocation_pool_end = module.params['allocation_pool_end']
    network = netaddr.IPNetwork(cidr)
    result = {
        'changed': False,
    }
    object_name = "%s_ips" % net_name
    net_allocations = allocations.setdefault(object_name, {})
    invalid_allocations = {hn: ip for hn, ip in net_allocations.items()
                           if netaddr.IPAddress(ip) not in network}
    if invalid_allocations:
        module.fail_json(msg="Found invalid existing allocations in network %s: %s" %
            (network,
             ", ".join("%s: %s" % (hn, ip)
                       for hn, ip in invalid_allocations.items())))
    if hostname not in net_allocations:
        result['changed'] = True
        allocated_ips = netaddr.IPSet(net_allocations.values())
        if allocation_pool_start and allocation_pool_end:
            allocation_pool = netaddr.IPRange(allocation_pool_start, allocation_pool_end)
            allocation_pool = netaddr.IPSet(allocation_pool)
        else:
            allocation_pool = netaddr.IPSet([network])
            if network.prefixlen != 32:
                reserved_ips = [network.network, network.broadcast]
                allocation_pool -= netaddr.IPSet(reserved_ips)
        free_ips = allocation_pool - allocated_ips
        for free_cidr in free_ips.iter_cidrs():
            ip = free_cidr[0]
            break
        else:
            module.fail_json(msg="No unallocated IP addresses for %s in %s" % (hostname, net_name))
        free_ips.remove(ip)
        net_allocations[hostname] = str(ip)
    result['ip'] = net_allocations[hostname]
    return result


def allocate(module):
    """Allocate an IP address for a host, updating the allocation file."""
    allocations = read_allocations(module)
    result = update_allocation(module, allocations)
    if result['changed'] and not module.check_mode:
        write_allocations(module, allocations)
    return result


def main():
    module = AnsibleModule(
        argument_spec=dict(
            net_name=dict(required=True, type='str'),
            hostname=dict(required=True, type='str'),
            cidr=dict(required=True, type='str'),
            allocation_pool_start=dict(required=False, type='str'),
            allocation_pool_end=dict(required=False, type='str'),
            allocation_file=dict(required=True, type='str'),
        ),
        supports_check_mode=True,
    )

    # Fail if there were any exceptions when importing modules.
    if IMPORT_ERRORS:
        module.fail_json(msg="Import errors: %s" %
                         ", ".join([repr(e) for e in IMPORT_ERRORS]))

    try:
        results = allocate(module)
    except Exception as e:
        module.fail_json(msg="Failed to allocate IP address: %s" % repr(e))
    else:
        module.exit_json(**results)


if __name__ == '__main__':
    main()



# Copyright (c) 2023 StackHPC Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from ansible.plugins.action import ActionBase
import os
from collections import defaultdict
import pathlib

from wcmatch import glob

def _dedup(xs):
    # Deduplicate a list whilst maintaining order
    seen = set()
    result = []
    for x in xs:
        if x not in seen:
            seen.add(x)
            result.append(x)
    return result

class ConfigCollector(object):
    def __init__(self, include_globs, ignore_globs, destination, search_paths,
                 rules):
        # This variable groups together files in the search paths with
        # the same relative path, for example if the search paths were:
        # - {{ kayobe_config_env_path }}/
        # - {{ kayobe_config_path }}/
        # - {{ role_path }}/templates/
        # and one of the include_globs matched nova.conf. You'd end up
        # with the following files grouped together:
        # - {{ kayobe_env_path }}/etc/kolla/nova.conf
        # - {{ kayobe_config_path }}/etc/kolla/nova.conf
        # - {{ role_path }}/templates/etc/kolla/nova.conf
        # The key in the dictionary is the relative path of the file. The
        # value is a list of absolute paths. This gets populated by the
        # collect() method.
        self.files_in_source = defaultdict(list)
        # Set of files in destination. This is used to cleanup up files
        # from a previous run that are no longer templated. Before any templating
        # this variable is populated with all the files in the output directory.
        # Each file that would be templated in the current run will be removed
        # from this set as they are discovered.
        self.files_in_destination = set()
        # Determines which files are candidates for templating
        self.include_globs = include_globs
        # Some files are templated by external tasks. This is a list of files
        # to not clean up.
        self.ignore_globs = ignore_globs
        # Where the files are being templated to
        self.destination = destination
        # Where to search for the source files
        self.search_paths = search_paths

        # Rules to determine merging strategy when multiple files are found
        # with the same relative path. Lower priority numbers win.
        self.rules = sorted(rules, key=lambda d: d['priority'])

    def filter_files_in_destination(self):
        ignored = set()
        for f in self.files_in_destination:
            for item in self.ignore_globs:
                if not item["enabled"]:
                    continue
                if glob.globmatch(f, item["glob"], flags=glob.GLOBSTAR):
                    ignored.add(f)
        result = set(self.files_in_destination) - ignored
        return list(result)

    def _find_matching_rule(self, relative_path, sources):
        # First match wins
        for rule in self.rules:
            if not rule.get('enabled', True):
                continue
            glob_ = rule["glob"]
            if glob.globmatch(relative_path, glob_, flags=glob.GLOBSTAR):
                requires_merge = (rule["strategy"] in
                    ["merge_configs", "merge_yaml"])
                # Fallback to templating when there is only one source. This
                # allows you to have config files that template to invalid
                # yaml/ini. This was allowed prior to config merging so
                # improves backwards compatibility.
                if requires_merge and len(sources) == 1:
                    # The rule can be used again to match a different file
                    # so don't modify in place.
                    rule = rule.copy()
                    rule["strategy"] = 'template'
                    # Strip parameters as they may not be compatible with
                    # template module.
                    rule['params'] = {}
                return rule

    def partition_into_actions(self):
        actions = {
            "merge_yaml": [],
            "merge_configs": [],
            "template": [],
            "copy": [],
            "concat": [],
            "create_dir": [],
            "delete": []
        }
        missing_directories = set()
        files_to_delete = self.filter_files_in_destination()

        # Convert to absolute paths
        files_to_delete = {
            os.path.join(self.destination, x) for x in files_to_delete
        }

        for relative_path, sources in self.files_in_source.items():
            found_match = False
            destination = os.path.join(self.destination, relative_path)
            # Don't delete any files we are templating
            files_to_delete.discard(destination)

            dirname = os.path.dirname(destination)
            if not os.path.exists(dirname):
                missing_directories.add(dirname)

            sources = map(os.path.realpath, sources)
            sources = _dedup(sources)

            rule = self._find_matching_rule(relative_path, sources)

            if not rule:
                continue

            if rule["strategy"] == 'copy':
                copy = {
                    "src": sources[-1],
                    "dest": destination,
                    "params": rule.get('params', [])
                }
                actions["copy"].append(copy)
                continue

            if rule["strategy"] == "merge_yaml":
                merge_yaml = {
                    "sources": sources,
                    "dest": destination,
                    "params": rule.get('params', [])
                }
                actions["merge_yaml"].append(merge_yaml)
                continue

            if rule["strategy"] == "merge_configs":
                merge_configs = {
                    "sources": sources,
                    "dest": destination,
                    "params": rule.get('params', [])
                }
                actions["merge_configs"].append(merge_configs)
                continue

            if rule["strategy"] == "concat":
                concat = {
                    "sources": sources,
                    "dest": destination,
                    "params": rule.get('params', [])
                }
                actions["concat"].append(concat)
                continue

            if rule["strategy"] == "template":
                template = {
                    "src": sources[-1],
                    "dest": destination,
                    "params": rule.get('params', [])
                }
                actions["template"].append(template)
                continue

        actions["create_dir"] = list(missing_directories)
        # Sort by length so that subdirectories are created after the parent
        actions["create_dir"].sort(key=len)

        actions["delete"] = list(files_to_delete)
        return actions

    def collect(self):
        for item in self.include_globs:
            self._collect_source(item)
            self._collect_destination(item)

    def _collect_source(self, item):
        enabled = item.get("enabled", False)
        if not isinstance(enabled, bool):
            raise ValueError("Expecting a boolean: %s" % item)
        if not enabled:
            return
        for search_path in self.search_paths:
            abs_glob = os.path.join(search_path, item["glob"])
            files = glob.glob(abs_glob, flags=glob.GLOBSTAR)
            for abs_path in files:
                if not os.path.isfile(abs_path):
                    continue
                relative_path = os.path.relpath(abs_path, search_path)
                self.files_in_source[relative_path].append(abs_path)

    def _collect_destination(self, item):
        abs_glob = os.path.join(self.destination, item["glob"])
        files = glob.glob(abs_glob, flags=glob.GLOBSTAR)
        for abs_path in files:
            if not os.path.isfile(abs_path):
                continue
            relative_path = os.path.relpath(abs_path, self.destination)
            self.files_in_destination.add(relative_path)

class ActionModule(ActionBase):

    def run(self, tmp=None, task_vars=None):
        if task_vars is None:
            task_vars = dict()

        result = super(ActionModule, self).run(tmp, task_vars)

        # This class never changes anything. We only collect the extra config
        # files and group by action.
        result['changed'] = False

        args = self._task.args

        collector = ConfigCollector(
            destination=args.get("destination"),
            ignore_globs=args.get("ignore_globs"),
            include_globs=args.get("include_globs"),
            rules=args.get("rules"),
            search_paths=_dedup(args["search_paths"])
        )

        collector.collect()

        result.update(collector.partition_into_actions())

        return result



# Copyright (c) 2017 StackHPC Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import glob
import os
import unittest
from unittest import mock

import cliff.app
import cliff.commandmanager

from kayobe import ansible
from kayobe.cli import commands
from kayobe import utils


class TestApp(cliff.app.App):

    def __init__(self):
        super(TestApp, self).__init__(
            description='Test app',
            version='0.1',
            command_manager=cliff.commandmanager.CommandManager('kayobe.cli'))


class TestCase(unittest.TestCase):

    maxDiff = None

    @mock.patch.object(ansible, "install_galaxy_roles", autospec=True)
    @mock.patch.object(ansible, "install_galaxy_collections", autospec=True)
    @mock.patch.object(ansible, "passwords_yml_exists", autospec=True)
    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    def test_control_host_bootstrap(self, mock_run, mock_passwords,
                                    mock_install_collections,
                                    mock_install_roles):
        mock_passwords.return_value = False
        command = commands.ControlHostBootstrap(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args([])
        result = command.run(parsed_args)
        self.assertEqual(0, result)
        mock_install_roles.assert_called_once_with(parsed_args)
        mock_install_collections.assert_called_once_with(parsed_args)
        expected_calls = [
            mock.call(
                mock.ANY,
                [utils.get_data_files_path("ansible", "bootstrap.yml")],
                ignore_limit=True,
            ),
            mock.call(
                mock.ANY,
                [utils.get_data_files_path("ansible", "kolla-ansible.yml")],
                tags="install",
                ignore_limit=True,
                check=False,
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(ansible, "install_galaxy_roles", autospec=True)
    @mock.patch.object(ansible, "install_galaxy_collections", autospec=True)
    @mock.patch.object(ansible, "passwords_yml_exists", autospec=True)
    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    @mock.patch.object(commands.KollaAnsibleMixin,
                       "run_kolla_ansible_overcloud")
    def test_control_host_bootstrap_with_passwords(
            self, mock_kolla_run, mock_run, mock_passwords,
            mock_install_collections, mock_install_roles):
        mock_passwords.return_value = True
        command = commands.ControlHostBootstrap(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args([])
        result = command.run(parsed_args)
        self.assertEqual(0, result)
        mock_install_roles.assert_called_once_with(parsed_args)
        mock_install_collections.assert_called_once_with(parsed_args)
        expected_calls = [
            mock.call(
                mock.ANY,
                [utils.get_data_files_path("ansible", "bootstrap.yml")],
                ignore_limit=True
            ),
            mock.call(
                mock.ANY,
                [utils.get_data_files_path("ansible", "kolla-ansible.yml")],
                tags=None,
                ignore_limit=True,
                check=False,
            ),
            mock.call(
                mock.ANY,
                [utils.get_data_files_path("ansible", "public-openrc.yml")],
                ignore_limit=True
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

        expected_calls = [
            mock.call(
                mock.ANY,
                "post-deploy",
            )
        ]
        self.assertListEqual(expected_calls, mock_kolla_run.call_args_list)

    @mock.patch.object(ansible, "install_galaxy_roles", autospec=True)
    @mock.patch.object(ansible, "install_galaxy_collections", autospec=True)
    @mock.patch.object(ansible, "prune_galaxy_roles", autospec=True)
    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    def test_control_host_upgrade(self, mock_run, mock_prune,
                                  mock_install_roles,
                                  mock_install_collections):
        command = commands.ControlHostUpgrade(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args([])
        result = command.run(parsed_args)
        self.assertEqual(0, result)
        mock_install_roles.assert_called_once_with(parsed_args, force=True)
        mock_install_collections.assert_called_once_with(parsed_args,
                                                         force=True)
        mock_prune.assert_called_once_with(parsed_args)
        expected_calls = [
            mock.call(
                mock.ANY,
                [utils.get_data_files_path("ansible", "bootstrap.yml")],
                ignore_limit=True,
            ),
            mock.call(
                mock.ANY,
                [utils.get_data_files_path("ansible", "kolla-ansible.yml")],
                tags="install",
                ignore_limit=True,
                check=False,
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbook")
    def test_physical_network_configure(self, mock_run):
        command = commands.PhysicalNetworkConfigure(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args(["--group", "switches"])
        result = command.run(parsed_args)
        self.assertEqual(0, result)
        expected_calls = [
            mock.call(
                mock.ANY,
                utils.get_data_files_path("ansible", "physical-network.yml"),
                limit="switches",
                extra_vars={
                    "physical_network_display": False
                }
            )
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbook")
    def test_physical_network_configure_display(self, mock_run):
        command = commands.PhysicalNetworkConfigure(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args(["--group", "switches", "--display"])
        result = command.run(parsed_args)
        self.assertEqual(0, result)
        expected_calls = [
            mock.call(
                mock.ANY,
                utils.get_data_files_path("ansible", "physical-network.yml"),
                limit="switches",
                extra_vars={
                    "physical_network_display": True
                }
            )
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbook")
    def test_physical_network_configure_enable_disco(self, mock_run):
        command = commands.PhysicalNetworkConfigure(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args(
            ["--group", "switches", "--enable-discovery"])
        result = command.run(parsed_args)
        self.assertEqual(0, result)
        expected_calls = [
            mock.call(
                mock.ANY,
                utils.get_data_files_path("ansible", "physical-network.yml"),
                limit="switches",
                extra_vars={
                    "physical_network_display": False,
                    "physical_network_enable_discovery": True
                }
            )
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbook")
    def test_physical_network_configure_disable_disco(self, mock_run):
        command = commands.PhysicalNetworkConfigure(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args(
            ["--group", "switches", "--disable-discovery"])
        result = command.run(parsed_args)
        self.assertEqual(0, result)
        expected_calls = [
            mock.call(
                mock.ANY,
                utils.get_data_files_path("ansible", "physical-network.yml"),
                limit="switches",
                extra_vars={
                    "physical_network_display": False,
                    "physical_network_disable_discovery": True
                }
            )
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    def test_physical_network_configure_enable_disable_disco(self):
        command = commands.PhysicalNetworkConfigure(TestApp(), [])
        parser = command.get_parser("test")
        self.assertRaises(
            SystemExit,
            parser.parse_args,
            ["--group", "switches", "--enable-discovery",
             "--disable-discovery"])

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbook")
    def test_physical_network_configure_interface_limit(self, mock_run):
        command = commands.PhysicalNetworkConfigure(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args(
            ["--group", "switches", "--interface-limit", "eth0,eth1"])
        result = command.run(parsed_args)
        self.assertEqual(0, result)
        expected_calls = [
            mock.call(
                mock.ANY,
                utils.get_data_files_path("ansible", "physical-network.yml"),
                limit="switches",
                extra_vars={
                    "physical_network_display": False,
                    "physical_network_interface_limit": "eth0,eth1"
                }
            )
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbook")
    def test_physical_network_configure_interface_description_limit(
            self, mock_run):
        command = commands.PhysicalNetworkConfigure(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args(
            ["--group", "switches",
             "--interface-description-limit", "host1,host2"])
        result = command.run(parsed_args)
        self.assertEqual(0, result)
        expected_calls = [
            mock.call(
                mock.ANY,
                utils.get_data_files_path("ansible", "physical-network.yml"),
                limit="switches",
                extra_vars={
                    "physical_network_display": False,
                    "physical_network_interface_description_limit": (
                        "host1,host2")
                }
            )
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    def test_network_connectivity_check(self, mock_run):
        command = commands.NetworkConnectivityCheck(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args([])
        result = command.run(parsed_args)
        self.assertEqual(0, result)
        expected_calls = [
            mock.call(mock.ANY, [utils.get_data_files_path(
                "ansible", "network-connectivity.yml")]),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    def test_seed_hypervisor_host_configure(self, mock_run):
        command = commands.SeedHypervisorHostConfigure(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args([])

        result = command.run(parsed_args)
        self.assertEqual(0, result)

        expected_calls = [
            mock.call(
                mock.ANY,
                [utils.get_data_files_path("ansible", "ip-allocation.yml")],
                limit="seed-hypervisor",
            ),
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path(
                        "ansible", "seed-hypervisor-host-configure.yml"),
                ],
                limit="seed-hypervisor",
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    def test_seed_hypervisor_host_configure_wipe_disks(self, mock_run):
        command = commands.SeedHypervisorHostConfigure(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args(["--wipe-disks"])

        result = command.run(parsed_args)
        self.assertEqual(0, result)

        expected_calls = [
            mock.call(
                mock.ANY,
                [utils.get_data_files_path("ansible", "ip-allocation.yml")],
                limit="seed-hypervisor",
            ),
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path(
                        "ansible", "seed-hypervisor-host-configure.yml"),
                ],
                limit="seed-hypervisor",
                extra_vars={"wipe_disks": True},
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    def test_seed_hypervisor_host_command_run(self, mock_run):
        command = commands.SeedHypervisorHostCommandRun(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args(["--command", "ls -a",
                                         "--show-output"])

        result = command.run(parsed_args)
        self.assertEqual(0, result)

        expected_calls = [
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path("ansible",
                                              "host-command-run.yml"),
                ],
                limit="seed-hypervisor",
                extra_vars={
                    "host_command_to_run": utils.escape_jinja("ls -a"),
                    "show_output": True}
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    def test_seed_hypervisor_host_package_update_all(self, mock_run):
        command = commands.SeedHypervisorHostPackageUpdate(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args(["--packages", "*"])

        result = command.run(parsed_args)
        self.assertEqual(0, result)

        expected_calls = [
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path(
                        "ansible", "host-package-update.yml"),
                ],
                limit="seed-hypervisor",
                extra_vars={
                    "host_package_update_packages": "*",
                    "host_package_update_security": False,
                },
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    def test_seed_hypervisor_host_package_update_list(self, mock_run):
        command = commands.SeedHypervisorHostPackageUpdate(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args(["--packages", "p1,p2"])

        result = command.run(parsed_args)
        self.assertEqual(0, result)

        expected_calls = [
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path(
                        "ansible", "host-package-update.yml"),
                ],
                limit="seed-hypervisor",
                extra_vars={
                    "host_package_update_packages": "p1,p2",
                    "host_package_update_security": False,
                },
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    def test_seed_hypervisor_host_package_update_security(self, mock_run):
        command = commands.SeedHypervisorHostPackageUpdate(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args(["--packages", "*", "--security"])

        result = command.run(parsed_args)
        self.assertEqual(0, result)

        expected_calls = [
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path(
                        "ansible", "host-package-update.yml"),
                ],
                limit="seed-hypervisor",
                extra_vars={
                    "host_package_update_packages": "*",
                    "host_package_update_security": True,
                },
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    def test_seed_hypervisor_host_upgrade(self, mock_run):
        command = commands.SeedHypervisorHostUpgrade(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args([])

        result = command.run(parsed_args)
        self.assertEqual(0, result)

        expected_calls = [
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path(
                        "ansible", "kayobe-target-venv.yml"),
                ],
                limit="seed-hypervisor",
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    def test_seed_host_configure(self, mock_run):
        command = commands.SeedHostConfigure(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args([])

        result = command.run(parsed_args)
        self.assertEqual(0, result)

        expected_calls = [
            mock.call(
                mock.ANY,
                [utils.get_data_files_path("ansible", "ip-allocation.yml")],
                limit="seed",
            ),
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path(
                        "ansible", "seed-host-configure.yml"),
                ],
                limit="seed",
                extra_vars={"kayobe_action": "deploy"},
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    def test_seed_host_configure_wipe_disks(self, mock_run):
        command = commands.SeedHostConfigure(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args(["--wipe-disks"])

        result = command.run(parsed_args)
        self.assertEqual(0, result)

        expected_calls = [
            mock.call(
                mock.ANY,
                [utils.get_data_files_path("ansible", "ip-allocation.yml")],
                limit="seed",
            ),
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path(
                        "ansible", "seed-host-configure.yml"),
                ],
                limit="seed",
                extra_vars={"kayobe_action": "deploy", "wipe_disks": True},
            ),
        ]
        print(expected_calls)
        print(mock_run.call_args_list)
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    def test_seed_host_command_run(self, mock_run):
        command = commands.SeedHostCommandRun(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args(["--command", "ls -a",
                                         "--show-output"])

        result = command.run(parsed_args)
        self.assertEqual(0, result)

        expected_calls = [
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path("ansible",
                                              "host-command-run.yml"),
                ],
                limit="seed",
                extra_vars={
                    "host_command_to_run": utils.escape_jinja("ls -a"),
                    "show_output": True}
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    def test_seed_host_package_update_all(self, mock_run):
        command = commands.SeedHostPackageUpdate(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args(["--packages", "*"])

        result = command.run(parsed_args)
        self.assertEqual(0, result)

        expected_calls = [
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path(
                        "ansible", "host-package-update.yml"),
                ],
                limit="seed",
                extra_vars={
                    "host_package_update_packages": "*",
                    "host_package_update_security": False,
                },
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    def test_seed_host_package_update_list(self, mock_run):
        command = commands.SeedHostPackageUpdate(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args(["--packages", "p1,p2"])

        result = command.run(parsed_args)
        self.assertEqual(0, result)

        expected_calls = [
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path(
                        "ansible", "host-package-update.yml"),
                ],
                limit="seed",
                extra_vars={
                    "host_package_update_packages": "p1,p2",
                    "host_package_update_security": False,
                },
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    def test_seed_host_package_update_security(self, mock_run):
        command = commands.SeedHostPackageUpdate(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args(["--packages", "*", "--security"])

        result = command.run(parsed_args)
        self.assertEqual(0, result)

        expected_calls = [
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path(
                        "ansible", "host-package-update.yml"),
                ],
                limit="seed",
                extra_vars={
                    "host_package_update_packages": "*",
                    "host_package_update_security": True,
                },
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    def test_seed_host_upgrade(self, mock_run):
        command = commands.SeedHostUpgrade(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args([])

        result = command.run(parsed_args)
        self.assertEqual(0, result)

        expected_calls = [
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path(
                        "ansible", "seed-host-upgrade.yml"),
                ],
                limit="seed",
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    def test_seed_container_image_build(self, mock_run):
        command = commands.SeedContainerImageBuild(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args([])
        result = command.run(parsed_args)
        self.assertEqual(0, result)
        expected_calls = [
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path(
                        "ansible", "container-image-builders-check.yml"),
                    utils.get_data_files_path("ansible", "kolla-build.yml"),
                    utils.get_data_files_path(
                        "ansible", "container-image-build.yml")
                ],
                extra_vars={
                    "container_image_sets": (
                        "{{ seed_container_image_sets }}"),
                    "push_images": False,
                    "nocache": False
                }
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    def test_seed_container_image_build_with_regex(self, mock_run):
        command = commands.SeedContainerImageBuild(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args(["--push", "^regex1$", "^regex2$"])
        result = command.run(parsed_args)
        self.assertEqual(0, result)
        expected_calls = [
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path(
                        "ansible", "container-image-builders-check.yml"),
                    utils.get_data_files_path("ansible", "kolla-build.yml"),
                    utils.get_data_files_path(
                        "ansible", "container-image-build.yml")
                ],
                extra_vars={
                    "container_image_regexes": "^regex1$ ^regex2$",
                    "push_images": True,
                    "nocache": False
                }
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    def test_seed_container_image_build_with_nocache(self, mock_run):
        command = commands.SeedContainerImageBuild(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args(["--nocache"])
        result = command.run(parsed_args)
        self.assertEqual(0, result)
        expected_calls = [
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path(
                        "ansible", "container-image-builders-check.yml"),
                    utils.get_data_files_path("ansible", "kolla-build.yml"),
                    utils.get_data_files_path(
                        "ansible", "container-image-build.yml")
                ],
                extra_vars={
                    "container_image_sets": (
                        "{{ seed_container_image_sets }}"),
                    "push_images": False,
                    "nocache": True
                }
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    def test_seed_deployment_image_build(self, mock_run):
        command = commands.SeedDeploymentImageBuild(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args([])

        result = command.run(parsed_args)
        self.assertEqual(0, result)

        expected_calls = [
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path("ansible", "seed-ipa-build.yml"),
                ],
                extra_vars={},
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    def test_seed_deployment_image_build_force_rebuild(self, mock_run):
        command = commands.SeedDeploymentImageBuild(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args(["--force-rebuild"])

        result = command.run(parsed_args)
        self.assertEqual(0, result)

        expected_calls = [
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path("ansible", "seed-ipa-build.yml"),
                ],
                extra_vars={"ipa_image_force_rebuild": True},
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    @mock.patch.object(commands.KollaAnsibleMixin,
                       "run_kolla_ansible_seed")
    def test_seed_service_deploy(self, mock_kolla_run, mock_run):
        command = commands.SeedServiceDeploy(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args([])

        result = command.run(parsed_args)
        self.assertEqual(0, result)

        expected_calls = [
            mock.call(
                mock.ANY,
                [utils.get_data_files_path("ansible", "seed-manage-containers.yml")],  # noqa
                extra_vars={'kayobe_action': 'deploy'}
            ),
            mock.call(
                mock.ANY,
                [utils.get_data_files_path("ansible", "kolla-ansible.yml")],
                tags="config",
                ignore_limit=True,
                check=False,
            ),
            mock.call(
                mock.ANY,
                [utils.get_data_files_path("ansible", "kolla-bifrost.yml")],
                ignore_limit=True,
                check=False,
            ),
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path(
                        "ansible", "seed-credentials.yml"),
                    utils.get_data_files_path(
                        "ansible", "seed-introspection-rules.yml"),
                    utils.get_data_files_path(
                        "ansible", "dell-switch-bmp.yml"),
                ],
                extra_vars={'kayobe_action': 'deploy'}
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

        expected_calls = [
            mock.call(
                mock.ANY,
                "deploy-bifrost",
            ),
        ]
        self.assertListEqual(expected_calls, mock_kolla_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    @mock.patch.object(commands.KollaAnsibleMixin,
                       "run_kolla_ansible_seed")
    def test_seed_service_upgrade(self, mock_kolla_run, mock_run):
        command = commands.SeedServiceUpgrade(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args([])

        result = command.run(parsed_args)
        self.assertEqual(0, result)

        expected_calls = [
            mock.call(
                mock.ANY,
                [utils.get_data_files_path("ansible", "seed-manage-containers.yml")],  # noqa
                extra_vars={'kayobe_action': 'deploy'}
            ),
            mock.call(
                mock.ANY,
                [utils.get_data_files_path("ansible", "kolla-ansible.yml")],
                tags="config",
                ignore_limit=True,
                check=False,
            ),
            mock.call(
                mock.ANY,
                [utils.get_data_files_path("ansible", "kolla-bifrost.yml")],
                ignore_limit=True,
                check=False,
            ),
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path("ansible",
                                              "seed-service-upgrade-prep.yml")
                ],
            ),
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path(
                        "ansible", "seed-credentials.yml"),
                    utils.get_data_files_path(
                        "ansible",
                        "seed-introspection-rules.yml"),
                    utils.get_data_files_path(
                        "ansible",
                        "dell-switch-bmp.yml"),
                ],
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

        expected_calls = [
            mock.call(
                mock.ANY,
                "upgrade-bifrost",
            ),
        ]
        self.assertListEqual(expected_calls, mock_kolla_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbook")
    def test_infra_vm_provision(self, mock_run):
        command = commands.InfraVMProvision(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args([])

        result = command.run(parsed_args)
        self.assertEqual(0, result)

        expected_calls = [
            mock.call(
                mock.ANY,
                utils.get_data_files_path(
                    "ansible", "ip-allocation.yml"),
                limit="infra-vms"
            ),
            mock.call(
                mock.ANY,
                utils.get_data_files_path(
                    "ansible", "infra-vm-provision.yml"),
                ignore_limit=True,
                extra_vars={'infra_vm_limit': 'infra-vms'}
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbook")
    def test_infra_vm_deprovision(self, mock_run):
        command = commands.InfraVMDeprovision(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args([])

        result = command.run(parsed_args)
        self.assertEqual(0, result)

        expected_calls = [
            mock.call(
                mock.ANY,
                utils.get_data_files_path(
                    "ansible", "infra-vm-deprovision.yml"),
                ignore_limit=True,
                extra_vars={'infra_vm_limit': 'infra-vms'}
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    def test_infra_vm_host_configure(self, mock_run):
        command = commands.InfraVMHostConfigure(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args([])

        result = command.run(parsed_args)
        self.assertEqual(0, result)

        expected_calls = [
            mock.call(
                mock.ANY,
                [utils.get_data_files_path("ansible", "ip-allocation.yml")],
                limit="infra-vms",
            ),
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path(
                        "ansible", "infra-vm-host-configure.yml"),
                ],
                limit="infra-vms",
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    def test_infra_vm_host_configure_wipe_disks(self, mock_run):
        command = commands.InfraVMHostConfigure(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args(["--wipe-disks"])

        result = command.run(parsed_args)
        self.assertEqual(0, result)

        expected_calls = [
            mock.call(
                mock.ANY,
                [utils.get_data_files_path("ansible", "ip-allocation.yml")],
                limit="infra-vms",
            ),
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path(
                        "ansible", "infra-vm-host-configure.yml"),
                ],
                limit="infra-vms",
                extra_vars={"wipe_disks": True},
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    def test_infra_vm_host_upgrade(self, mock_run):
        command = commands.InfraVMHostUpgrade(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args([])

        result = command.run(parsed_args)
        self.assertEqual(0, result)

        expected_calls = [
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path(
                        "ansible", "kayobe-target-venv.yml"),
                ],
                limit="infra-vms",
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    def test_infra_vm_host_command_run(self, mock_run):
        command = commands.InfraVMHostCommandRun(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args(["--command", "ls -a",
                                         "--show-output"])

        result = command.run(parsed_args)
        self.assertEqual(0, result)

        expected_calls = [
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path("ansible",
                                              "host-command-run.yml"),
                ],
                limit="infra-vms",
                extra_vars={
                    "host_command_to_run": utils.escape_jinja("ls -a"),
                    "show_output": True}
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    def test_infra_vm_host_package_update_all(self, mock_run):
        command = commands.InfraVMHostPackageUpdate(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args(["--packages", "*"])

        result = command.run(parsed_args)
        self.assertEqual(0, result)

        expected_calls = [
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path(
                        "ansible", "host-package-update.yml"),
                ],
                limit="infra-vms",
                extra_vars={
                    "host_package_update_packages": "*",
                    "host_package_update_security": False,
                },
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbook")
    def test_infra_vm_service_deploy(self, mock_run):
        command = commands.InfraVMServiceDeploy(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args([])

        result = command.run(parsed_args)
        self.assertEqual(0, result)

        expected_calls = []
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbook")
    def test_overcloud_inventory_discover(self, mock_run_one, mock_run):
        command = commands.OvercloudInventoryDiscover(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args([])

        result = command.run(parsed_args)
        self.assertEqual(0, result)

        expected_calls = [
            mock.call(
                mock.ANY,
                utils.get_data_files_path(
                    "ansible", "overcloud-inventory-discover.yml"),
            ),
            mock.call(
                mock.ANY,
                utils.get_data_files_path("ansible", "ip-allocation.yml"),
                limit="overcloud",
            ),
        ]
        self.assertListEqual(expected_calls, mock_run_one.call_args_list)

        expected_calls = [
            mock.call(
                mock.ANY,
                [utils.get_data_files_path("ansible", "kolla-ansible.yml")],
                tags="config",
                ignore_limit=True,
                check=False,
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    def test_overcloud_hardware_inspect(self, mock_run):
        command = commands.OvercloudHardwareInspect(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args([])

        result = command.run(parsed_args)
        self.assertEqual(0, result)

        expected_calls = [
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path(
                        "ansible", "kolla-bifrost-hostvars.yml"),
                    utils.get_data_files_path(
                        "ansible", "overcloud-hardware-inspect.yml"),
                ],
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    def test_overcloud_provision(self, mock_run):
        command = commands.OvercloudProvision(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args([])

        result = command.run(parsed_args)
        self.assertEqual(0, result)

        expected_calls = [
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path(
                        "ansible", "kolla-bifrost-hostvars.yml"),
                    utils.get_data_files_path(
                        "ansible", "overcloud-provision.yml"),
                ],
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    def test_overcloud_deprovision(self, mock_run):
        command = commands.OvercloudDeprovision(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args([])

        result = command.run(parsed_args)
        self.assertEqual(0, result)

        expected_calls = [
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path(
                        "ansible", "kolla-bifrost-hostvars.yml"),
                    utils.get_data_files_path(
                        "ansible", "overcloud-deprovision.yml"),
                ],
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    @mock.patch.object(commands.KollaAnsibleMixin,
                       "run_kolla_ansible_overcloud")
    def test_overcloud_facts_gather(self, mock_kolla_run, mock_run):
        command = commands.OvercloudFactsGather(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args([])

        result = command.run(parsed_args)
        self.assertEqual(0, result)

        expected_calls = [
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path(
                        "ansible", "overcloud-facts-gather.yml"),
                ],
            ),
            mock.call(
                mock.ANY,
                [utils.get_data_files_path("ansible", "kolla-ansible.yml")],
                tags="config",
                ignore_limit=True,
                check=False,
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

        expected_calls = [
            mock.call(
                mock.ANY,
                "gather-facts"
            ),
        ]
        self.assertListEqual(expected_calls, mock_kolla_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    def test_overcloud_host_configure(self, mock_run):
        command = commands.OvercloudHostConfigure(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args([])

        result = command.run(parsed_args)
        self.assertEqual(0, result)

        expected_calls = [
            mock.call(
                mock.ANY,
                [utils.get_data_files_path("ansible", "ip-allocation.yml")],
                limit="overcloud",
            ),
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path(
                        "ansible", "overcloud-host-configure.yml"),
                ],
                limit="overcloud",
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    def test_overcloud_host_configure_wipe_disks(self, mock_run):
        command = commands.OvercloudHostConfigure(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args(["--wipe-disks"])

        result = command.run(parsed_args)
        self.assertEqual(0, result)

        expected_calls = [
            mock.call(
                mock.ANY,
                [utils.get_data_files_path("ansible", "ip-allocation.yml")],
                limit="overcloud",
            ),
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path(
                        "ansible", "overcloud-host-configure.yml"),
                ],
                limit="overcloud",
                extra_vars={"wipe_disks": True},
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    def test_overcloud_host_command_run(self, mock_run):
        command = commands.OvercloudHostCommandRun(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args(["--command", "ls -a",
                                         "--show-output"])

        result = command.run(parsed_args)
        self.assertEqual(0, result)

        expected_calls = [
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path("ansible",
                                              "host-command-run.yml"),
                ],
                limit="overcloud",
                extra_vars={
                    "host_command_to_run": utils.escape_jinja("ls -a"),
                    "show_output": True}
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    def test_overcloud_host_package_update_all(self, mock_run):
        command = commands.OvercloudHostPackageUpdate(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args(["--packages", "*"])

        result = command.run(parsed_args)
        self.assertEqual(0, result)

        expected_calls = [
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path(
                        "ansible", "host-package-update.yml"),
                ],
                limit="overcloud",
                extra_vars={
                    "host_package_update_packages": "*",
                    "host_package_update_security": False,
                },
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    def test_overcloud_host_package_update_list(self, mock_run):
        command = commands.OvercloudHostPackageUpdate(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args(["--packages", "p1,p2"])

        result = command.run(parsed_args)
        self.assertEqual(0, result)

        expected_calls = [
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path(
                        "ansible", "host-package-update.yml"),
                ],
                limit="overcloud",
                extra_vars={
                    "host_package_update_packages": "p1,p2",
                    "host_package_update_security": False,
                },
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    def test_overcloud_host_package_update_security(self, mock_run):
        command = commands.OvercloudHostPackageUpdate(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args(["--packages", "*", "--security"])

        result = command.run(parsed_args)
        self.assertEqual(0, result)

        expected_calls = [
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path(
                        "ansible", "host-package-update.yml"),
                ],
                limit="overcloud",
                extra_vars={
                    "host_package_update_packages": "*",
                    "host_package_update_security": True,
                },
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    def test_overcloud_host_upgrade(self, mock_run):
        command = commands.OvercloudHostUpgrade(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args([])

        result = command.run(parsed_args)
        self.assertEqual(0, result)

        expected_calls = [
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path(
                        "ansible", "overcloud-host-upgrade.yml"),
                ],
                limit="overcloud",
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    @mock.patch.object(commands.KollaAnsibleMixin,
                       "run_kolla_ansible_overcloud")
    def test_overcloud_database_backup(self, mock_kolla_run, mock_run):
        command = commands.OvercloudDatabaseBackup(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args([])
        result = command.run(parsed_args)
        self.assertEqual(0, result)

        expected_calls = [
            mock.call(
                mock.ANY,
                [utils.get_data_files_path("ansible", "kolla-ansible.yml")],
                tags="config",
                ignore_limit=True,
                check=False,
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

        expected_calls = [
            mock.call(
                mock.ANY,
                "mariadb-backup",
                extra_args=[]
            ),
        ]
        self.assertListEqual(expected_calls, mock_kolla_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    @mock.patch.object(commands.KollaAnsibleMixin,
                       "run_kolla_ansible_overcloud")
    def test_overcloud_database_backup_incremental(self, mock_kolla_run,
                                                   mock_run):
        command = commands.OvercloudDatabaseBackup(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args(["--incremental"])
        result = command.run(parsed_args)
        self.assertEqual(0, result)

        expected_calls = [
            mock.call(
                mock.ANY,
                [utils.get_data_files_path("ansible", "kolla-ansible.yml")],
                tags="config",
                ignore_limit=True,
                check=False,
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

        expected_calls = [
            mock.call(
                mock.ANY,
                "mariadb-backup",
                extra_args=["--incremental"]
            ),
        ]
        self.assertListEqual(expected_calls, mock_kolla_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    @mock.patch.object(commands.KollaAnsibleMixin,
                       "run_kolla_ansible_overcloud")
    def test_overcloud_database_recover(self, mock_kolla_run, mock_run):
        command = commands.OvercloudDatabaseRecover(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args([])
        result = command.run(parsed_args)
        self.assertEqual(0, result)

        expected_calls = [
            mock.call(
                mock.ANY,
                [utils.get_data_files_path("ansible", "kolla-ansible.yml")],
                tags="config",
                ignore_limit=True,
                check=False,
            ),
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path("ansible",
                                              "kolla-openstack.yml"),
                ],
                ignore_limit=True,
                check=False,
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

        expected_calls = [
            mock.call(
                mock.ANY,
                "mariadb-recovery",
                extra_vars={}
            ),
        ]
        self.assertListEqual(expected_calls, mock_kolla_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    @mock.patch.object(commands.KollaAnsibleMixin,
                       "run_kolla_ansible_overcloud")
    def test_overcloud_database_recover_force_host(self, mock_kolla_run,
                                                   mock_run):
        command = commands.OvercloudDatabaseRecover(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args(["--force-recovery-host", "foo"])
        result = command.run(parsed_args)
        self.assertEqual(0, result)

        expected_calls = [
            mock.call(
                mock.ANY,
                [utils.get_data_files_path("ansible", "kolla-ansible.yml")],
                tags="config",
                ignore_limit=True,
                check=False,
            ),
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path("ansible",
                                              "kolla-openstack.yml"),
                ],
                ignore_limit=True,
                check=False,
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

        expected_calls = [
            mock.call(
                mock.ANY,
                "mariadb-recovery",
                extra_vars={
                    "mariadb_recover_inventory_name": "foo"
                }
            ),
        ]
        self.assertListEqual(expected_calls, mock_kolla_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    def test_overcloud_service_configuration_save(self, mock_run):
        command = commands.OvercloudServiceConfigurationSave(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args([])
        result = command.run(parsed_args)
        self.assertEqual(0, result)
        expected_calls = [
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path(
                        "ansible", "overcloud-service-config-save.yml"),
                ],
                extra_vars={}
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    @mock.patch.object(commands.KollaAnsibleMixin,
                       "run_kolla_ansible_overcloud")
    def test_overcloud_service_deploy(self, mock_kolla_run, mock_run):
        command = commands.OvercloudServiceDeploy(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args([])

        result = command.run(parsed_args)
        self.assertEqual(0, result)

        expected_calls = [
            mock.call(
                mock.ANY,
                [utils.get_data_files_path("ansible", "kolla-ansible.yml")],
                ignore_limit=True,
                tags="config",
                check=False,
            ),
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path("ansible",
                                              "kolla-openstack.yml"),
                ],
                ignore_limit=True,
                check=False,
            ),
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path("ansible",
                                              "overcloud-extras.yml"),
                ],
                limit="overcloud",
                extra_vars={
                    "kayobe_action": "deploy",
                },
            ),
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path("ansible", "public-openrc.yml"),
                ],
                ignore_limit=True,
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

        expected_calls = [
            mock.call(
                mock.ANY,
                "prechecks",
            ),
            mock.call(
                mock.ANY,
                "deploy",
            ),
            mock.call(
                mock.ANY,
                "post-deploy",
            ),
        ]
        self.assertListEqual(expected_calls, mock_kolla_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    @mock.patch.object(commands.KollaAnsibleMixin,
                       "run_kolla_ansible_overcloud")
    def test_overcloud_service_deploy_containers(self, mock_kolla_run,
                                                 mock_run):
        command = commands.OvercloudServiceDeployContainers(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args([])

        result = command.run(parsed_args)
        self.assertEqual(0, result)

        expected_calls = [
            mock.call(
                mock.ANY,
                [utils.get_data_files_path("ansible", "kolla-ansible.yml")],
                ignore_limit=True,
                tags="config",
                check=False,
            ),
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path("ansible",
                                              "kolla-openstack.yml"),
                ],
                ignore_limit=True,
                check=False,
            ),
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path("ansible",
                                              "overcloud-extras.yml"),
                ],
                limit="overcloud",
                extra_vars={
                    "kayobe_action": "deploy",
                },
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

        expected_calls = [
            mock.call(
                mock.ANY,
                "prechecks",
            ),
            mock.call(
                mock.ANY,
                "manage-containers",
            ),
        ]
        self.assertListEqual(expected_calls, mock_kolla_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    @mock.patch.object(commands.KollaAnsibleMixin,
                       "run_kolla_ansible_overcloud")
    def test_overcloud_service_prechecks(self, mock_kolla_run, mock_run):
        command = commands.OvercloudServicePrechecks(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args([])

        result = command.run(parsed_args)
        self.assertEqual(0, result)

        expected_calls = [
            mock.call(
                mock.ANY,
                [utils.get_data_files_path("ansible", "kolla-ansible.yml")],
                ignore_limit=True,
                tags="config",
                check=False,
            ),
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path("ansible",
                                              "kolla-openstack.yml"),
                ],
                ignore_limit=True,
                check=False,
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

        expected_calls = [
            mock.call(
                mock.ANY,
                "prechecks",
            ),
        ]
        self.assertListEqual(expected_calls, mock_kolla_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    @mock.patch.object(commands.KollaAnsibleMixin,
                       "run_kolla_ansible_overcloud")
    def test_overcloud_service_reconfigure(self, mock_kolla_run, mock_run):
        command = commands.OvercloudServiceReconfigure(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args([])

        result = command.run(parsed_args)
        self.assertEqual(0, result)

        expected_calls = [
            mock.call(
                mock.ANY,
                [utils.get_data_files_path("ansible", "kolla-ansible.yml")],
                ignore_limit=True,
                tags="config",
                check=False,
            ),
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path("ansible",
                                              "kolla-openstack.yml"),
                ],
                ignore_limit=True,
                check=False,
            ),
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path("ansible",
                                              "overcloud-extras.yml"),
                ],
                limit="overcloud",
                extra_vars={
                    "kayobe_action": "reconfigure",
                },
            ),
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path("ansible", "public-openrc.yml"),
                ],
                ignore_limit=True,
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

        expected_calls = [
            mock.call(
                mock.ANY,
                "prechecks",
            ),
            mock.call(
                mock.ANY,
                "reconfigure",
            ),
            mock.call(
                mock.ANY,
                "post-deploy",
            ),
        ]
        self.assertListEqual(expected_calls, mock_kolla_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    @mock.patch.object(commands.KollaAnsibleMixin,
                       "run_kolla_ansible_overcloud")
    def test_overcloud_service_stop(self, mock_kolla_run, mock_run):
        command = commands.OvercloudServiceStop(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args(["--yes-i-really-really-mean-it"])

        result = command.run(parsed_args)
        self.assertEqual(0, result)

        expected_calls = [
            mock.call(
                mock.ANY,
                [utils.get_data_files_path("ansible", "kolla-ansible.yml")],
                ignore_limit=True,
                tags="config",
                check=False,
            ),
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path("ansible",
                                              "kolla-openstack.yml"),
                ],
                ignore_limit=True,
                check=False,
            ),
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path("ansible",
                                              "overcloud-extras.yml"),
                ],
                limit="overcloud",
                extra_vars={
                    "kayobe_action": "stop",
                },
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

        expected_calls = [
            mock.call(
                mock.ANY,
                "stop",
                extra_args=["--yes-i-really-really-mean-it"],
            ),
        ]
        self.assertListEqual(expected_calls, mock_kolla_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    @mock.patch.object(commands.KollaAnsibleMixin,
                       "run_kolla_ansible_overcloud")
    def test_overcloud_service_stop_no_disclaimer(self, mock_kolla_run,
                                                  mock_run):
        command = commands.OvercloudServiceStop(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args([])

        self.assertRaises(
            SystemExit,
            command.run, parsed_args)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    @mock.patch.object(commands.KollaAnsibleMixin,
                       "run_kolla_ansible_overcloud")
    def test_overcloud_service_upgrade(self, mock_kolla_run, mock_run):
        command = commands.OvercloudServiceUpgrade(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args([])

        result = command.run(parsed_args)
        self.assertEqual(0, result)

        expected_calls = [
            mock.call(
                mock.ANY,
                [utils.get_data_files_path("ansible", "kolla-ansible.yml")],
                ignore_limit=True,
                tags=None,
                check=False,
            ),
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path("ansible",
                                              "kolla-openstack.yml"),
                ],
                ignore_limit=True,
                check=False,
            ),
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path("ansible",
                                              "overcloud-extras.yml"),
                ],
                limit="overcloud",
                extra_vars={
                    "kayobe_action": "upgrade",
                }
            ),
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path("ansible",
                                              "public-openrc.yml"),
                ],
                ignore_limit=True,
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

        expected_calls = [
            mock.call(
                mock.ANY,
                "prechecks"
            ),
            mock.call(
                mock.ANY,
                "upgrade"
            ),
            mock.call(
                mock.ANY,
                "post-deploy"
            ),
        ]
        self.assertListEqual(expected_calls, mock_kolla_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    def test_overcloud_service_configuration_save_args(self, mock_run):
        command = commands.OvercloudServiceConfigurationSave(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args([
            "--exclude", "exclude1,exclude2",
            "--include", "include1,include2",
            "--node-config-dir", "/path/to/config",
            "--output-dir", "/path/to/output",
        ])
        result = command.run(parsed_args)
        self.assertEqual(0, result)
        expected_calls = [
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path(
                        "ansible", "overcloud-service-config-save.yml"),
                ],
                extra_vars={
                    "exclude_patterns": "exclude1,exclude2",
                    "include_patterns": "include1,include2",
                    "config_save_path": "/path/to/output",
                    "node_config_directory": "/path/to/config",
                }
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    def test_overcloud_container_image_build(self, mock_run):
        command = commands.OvercloudContainerImageBuild(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args([])
        result = command.run(parsed_args)
        self.assertEqual(0, result)
        expected_calls = [
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path(
                        "ansible", "container-image-builders-check.yml"),
                    utils.get_data_files_path("ansible", "kolla-build.yml"),
                    utils.get_data_files_path(
                        "ansible", "container-image-build.yml")
                ],
                extra_vars={
                    "container_image_sets": (
                        "{{ overcloud_container_image_sets }}"),
                    "push_images": False,
                    "nocache": False
                }
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    def test_overcloud_container_image_build_with_regex(self, mock_run):
        command = commands.OvercloudContainerImageBuild(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args(["--push", "^regex1$", "^regex2$"])
        result = command.run(parsed_args)
        self.assertEqual(0, result)
        expected_calls = [
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path(
                        "ansible", "container-image-builders-check.yml"),
                    utils.get_data_files_path("ansible", "kolla-build.yml"),
                    utils.get_data_files_path(
                        "ansible", "container-image-build.yml")
                ],
                extra_vars={
                    "container_image_regexes": "^regex1$ ^regex2$",
                    "push_images": True,
                    "nocache": False
                }
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    def test_overcloud_host_image_build(self, mock_run):
        command = commands.OvercloudHostImageBuild(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args([])

        result = command.run(parsed_args)
        self.assertEqual(0, result)

        expected_calls = [
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path(
                        "ansible", "overcloud-host-image-build.yml"),
                ],
                extra_vars={},
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    def test_overcloud_host_image_build_force_rebuild(self, mock_run):
        command = commands.OvercloudHostImageBuild(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args(["--force-rebuild"])

        result = command.run(parsed_args)
        self.assertEqual(0, result)

        expected_calls = [
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path(
                        "ansible", "overcloud-host-image-build.yml"),
                ],
                extra_vars={"overcloud_host_image_force_rebuild": True},
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    def test_overcloud_deployment_image_build(self, mock_run):
        command = commands.OvercloudDeploymentImageBuild(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args([])

        result = command.run(parsed_args)
        self.assertEqual(0, result)

        expected_calls = [
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path(
                        "ansible", "overcloud-ipa-build.yml"),
                ],
                extra_vars={},
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    def test_overcloud_deployment_image_build_force_rebuild(self, mock_run):
        command = commands.OvercloudDeploymentImageBuild(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args(["--force-rebuild"])

        result = command.run(parsed_args)
        self.assertEqual(0, result)

        expected_calls = [
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path(
                        "ansible", "overcloud-ipa-build.yml"),
                ],
                extra_vars={"ipa_image_force_rebuild": True},
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    def test_overcloud_post_configure(self, mock_run):
        command = commands.OvercloudPostConfigure(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args([])

        result = command.run(parsed_args)
        self.assertEqual(0, result)

        expected_calls = [
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path(
                        "ansible", "overcloud-ipa-images.yml"),
                    utils.get_data_files_path(
                        "ansible", "overcloud-introspection-rules.yml"),
                    utils.get_data_files_path("ansible", "overcloud-introspection-rules-dell-lldp-workaround.yml"),  # noqa
                    utils.get_data_files_path("ansible", "provision-net.yml"),
                    utils.get_data_files_path(
                        "ansible", "baremetal-compute-serial-console-post-config.yml"),  # noqa
                ],
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    def test_overcloud_swift_rings_generate(self, mock_run):
        command = commands.OvercloudSwiftRingsGenerate(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args([])

        result = command.run(parsed_args)
        self.assertEqual(0, result)

        expected_calls = [
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path("ansible", "swift-rings.yml"),
                ],
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    def test_baremetal_compute_register(self, mock_run):
        command = commands.BaremetalComputeRegister(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args([])
        result = command.run(parsed_args)
        self.assertEqual(0, result)
        expected_calls = [
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path(
                        "ansible", "baremetal-compute-register.yml"),
                ],
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    def test_baremetal_compute_inspect(self, mock_run):
        command = commands.BaremetalComputeInspect(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args([])
        result = command.run(parsed_args)
        self.assertEqual(0, result)
        expected_calls = [
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path(
                        "ansible", "baremetal-compute-inspect.yml"),
                ],
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    def test_baremetal_compute_manage(self, mock_run):
        command = commands.BaremetalComputeManage(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args([])
        result = command.run(parsed_args)
        self.assertEqual(0, result)
        expected_calls = [
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path(
                        "ansible", "baremetal-compute-manage.yml"),
                ],
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    def test_baremetal_compute_provide(self, mock_run):
        command = commands.BaremetalComputeProvide(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args([])
        result = command.run(parsed_args)
        self.assertEqual(0, result)
        expected_calls = [
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path(
                        "ansible", "baremetal-compute-provide.yml"),
                ],
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    def test_baremetal_compute_rename(self, mock_run):
        command = commands.BaremetalComputeRename(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args([])
        result = command.run(parsed_args)
        self.assertEqual(0, result)
        expected_calls = [
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path(
                        "ansible", "baremetal-compute-rename.yml"),
                ],
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    def test_baremetal_compute_serial_console_enable(self, mock_run):
        command = commands.BaremetalComputeSerialConsoleEnable(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args([])
        result = command.run(parsed_args)
        self.assertEqual(0, result)
        expected_calls = [
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path(
                        "ansible", "baremetal-compute-serial-console.yml"),

                ],
                extra_vars={
                    "cmd": "enable",
                }
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    def test_baremetal_compute_serial_console_enable_with_limit(self,
                                                                mock_run):
        command = commands.BaremetalComputeSerialConsoleEnable(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args(["--baremetal-compute-limit",
                                         "sand-6-1"])
        result = command.run(parsed_args)
        self.assertEqual(0, result)
        expected_calls = [
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path(
                        "ansible", "baremetal-compute-serial-console.yml"),

                ],
                extra_vars={
                    "cmd": "enable",
                    "console_compute_node_limit": "sand-6-1",
                }
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    def test_baremetal_compute_serial_console_disable(self, mock_run):
        command = commands.BaremetalComputeSerialConsoleDisable(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args([])
        result = command.run(parsed_args)
        self.assertEqual(0, result)
        expected_calls = [
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path(
                        "ansible", "baremetal-compute-serial-console.yml"),

                ],
                extra_vars={
                    "cmd": "disable",
                }
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    def test_baremetal_compute_serial_console_disable_with_limit(self,
                                                                 mock_run):
        command = commands.BaremetalComputeSerialConsoleDisable(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args(["--baremetal-compute-limit",
                                         "sand-6-1"])
        result = command.run(parsed_args)
        self.assertEqual(0, result)
        expected_calls = [
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path(
                        "ansible", "baremetal-compute-serial-console.yml"),

                ],
                extra_vars={
                    "cmd": "disable",
                    "console_compute_node_limit": "sand-6-1",
                }
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    def test_baremetal_compute_update_deployment_image(self, mock_run):
        command = commands.BaremetalComputeUpdateDeploymentImage(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args([])
        result = command.run(parsed_args)
        self.assertEqual(0, result)
        expected_calls = [
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path(
                        "ansible", "overcloud-ipa-images.yml"),
                ],
                extra_vars={
                    "ipa_images_update_ironic_nodes": True,
                }
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)

    @mock.patch.object(commands.KayobeAnsibleMixin,
                       "run_kayobe_playbooks")
    def test_baremetal_compute_update_deployment_image_with_limit(
            self, mock_run):
        command = commands.BaremetalComputeUpdateDeploymentImage(TestApp(), [])
        parser = command.get_parser("test")
        parsed_args = parser.parse_args(["--baremetal-compute-limit",
                                         "sand-6-1"])
        result = command.run(parsed_args)
        self.assertEqual(0, result)
        expected_calls = [
            mock.call(
                mock.ANY,
                [
                    utils.get_data_files_path(
                        "ansible", "overcloud-ipa-images.yml"),
                ],
                extra_vars={
                    "ipa_images_compute_node_limit": "sand-6-1",
                    "ipa_images_update_ironic_nodes": True,
                }
            ),
        ]
        self.assertListEqual(expected_calls, mock_run.call_args_list)


class TestHookDispatcher(unittest.TestCase):

    maxDiff = None

    @mock.patch.object(os.path, 'realpath')
    def test_hook_ordering(self, mock_realpath):
        mock_command = mock.MagicMock()
        dispatcher = commands.HookDispatcher(command=mock_command)
        dispatcher._find_hooks = mock.MagicMock()
        # Include multiple hook directories to show that they don't influence
        # the order.
        dispatcher._find_hooks.return_value = [
            "config/path/10-hook.yml",
            "config/path/5-hook.yml",
            "config/path/z-test-alphabetical.yml",
            "env/path/10-before-hook.yml",
            "env/path/5-multiple-dashes-in-name.yml",
            "env/path/no-prefix.yml"
        ]
        expected_result = [
            "config/path/5-hook.yml",
            "env/path/5-multiple-dashes-in-name.yml",
            "env/path/10-before-hook.yml",
            "config/path/10-hook.yml",
            "env/path/no-prefix.yml",
            "config/path/z-test-alphabetical.yml",
        ]
        mock_realpath.side_effect = lambda x: x
        actual = dispatcher.hooks(["config/path", "env/path"], "pre", None)
        self.assertListEqual(actual, expected_result)

    @mock.patch('kayobe.cli.commands.os.path')
    def test_hook_filter_all(self, mock_path):
        mock_command = mock.MagicMock()
        dispatcher = commands.HookDispatcher(command=mock_command)
        dispatcher._find_hooks = mock.MagicMock()
        dispatcher._find_hooks.return_value = [
            "5-hook.yml",
            "5-multiple-dashes-in-name.yml",
            "10-before-hook.yml",
            "10-hook.yml",
            "no-prefix.yml",
            "z-test-alphabetical.yml",
        ]
        mock_path.realpath.side_effect = lambda x: x
        actual = dispatcher.hooks(["config/path"], "pre", "all")
        self.assertListEqual(actual, [])

    @mock.patch('kayobe.cli.commands.os.path')
    def test_hook_filter_one(self, mock_path):
        mock_command = mock.MagicMock()
        dispatcher = commands.HookDispatcher(command=mock_command)
        dispatcher._find_hooks = mock.MagicMock()
        dispatcher._find_hooks.return_value = [
            "5-hook.yml",
            "5-multiple-dashes-in-name.yml",
            "10-before-hook.yml",
            "10-hook.yml",
            "no-prefix.yml",
            "z-test-alphabetical.yml",
        ]
        expected_result = [
            "5-hook.yml",
            "10-before-hook.yml",
            "10-hook.yml",
            "no-prefix.yml",
            "z-test-alphabetical.yml",
        ]
        mock_path.realpath.side_effect = lambda x: x
        actual = dispatcher.hooks(["config/path"], "pre",
                                  "5-multiple-dashes-in-name.yml")
        self.assertListEqual(actual, expected_result)

    @mock.patch.object(glob, 'glob')
    @mock.patch.object(os.path, 'exists')
    def test__find_hooks(self, mock_exists, mock_glob):
        mock_exists.return_value = True
        mock_command = mock.MagicMock()
        dispatcher = commands.HookDispatcher(command=mock_command)
        mock_glob.return_value = [
            "config/path/hooks/pre.d/1-hook.yml",
            "config/path/hooks/pre.d/5-hook.yml",
            "config/path/hooks/pre.d/10-hook.yml",
        ]
        expected_result = [
            "config/path/hooks/pre.d/1-hook.yml",
            "config/path/hooks/pre.d/10-hook.yml",
            "config/path/hooks/pre.d/5-hook.yml",
        ]
        actual = dispatcher._find_hooks(["config/path"], "pre")
        # Sort the result - it is not ordered at this stage.
        actual.sort()
        self.assertListEqual(actual, expected_result)

    @mock.patch.object(glob, 'glob')
    @mock.patch.object(os.path, 'exists')
    def test__find_hooks_with_env(self, mock_exists, mock_glob):
        mock_exists.return_value = True
        mock_command = mock.MagicMock()
        dispatcher = commands.HookDispatcher(command=mock_command)
        mock_glob.side_effect = [
            [
                "config/path/hooks/pre.d/all.yml",
                "config/path/hooks/pre.d/base-only.yml",
            ],
            [
                "env/path/hooks/pre.d/all.yml",
                "env/path/hooks/pre.d/env-only.yml",
            ]
        ]
        expected_result = [
            "config/path/hooks/pre.d/base-only.yml",
            "env/path/hooks/pre.d/all.yml",
            "env/path/hooks/pre.d/env-only.yml",
        ]
        actual = dispatcher._find_hooks(["config/path", "env/path"], "pre")
        # Sort the result - it is not ordered at this stage.
        actual.sort()
        self.assertListEqual(actual, expected_result)

    @mock.patch.object(glob, 'glob')
    @mock.patch.object(os.path, 'exists')
    def test__find_hooks_with_nested_envs(self, mock_exists, mock_glob):
        mock_exists.return_value = True
        mock_command = mock.MagicMock()
        dispatcher = commands.HookDispatcher(command=mock_command)
        mock_glob.side_effect = [
            [
                "config/path/hooks/pre.d/all.yml",
                "config/path/hooks/pre.d/base-only.yml",
                "config/path/hooks/pre.d/base-env1.yml",
                "config/path/hooks/pre.d/base-env2.yml",
            ],
            [
                "env1/path/hooks/pre.d/all.yml",
                "env1/path/hooks/pre.d/env1-only.yml",
                "env1/path/hooks/pre.d/base-env1.yml",
                "env1/path/hooks/pre.d/env1-env2.yml",
            ],
            [
                "env2/path/hooks/pre.d/all.yml",
                "env2/path/hooks/pre.d/env2-only.yml",
                "env2/path/hooks/pre.d/base-env2.yml",
                "env2/path/hooks/pre.d/env1-env2.yml",
            ]
        ]
        expected_result = [
            "config/path/hooks/pre.d/base-only.yml",
            "env1/path/hooks/pre.d/base-env1.yml",
            "env1/path/hooks/pre.d/env1-only.yml",
            "env2/path/hooks/pre.d/all.yml",
            "env2/path/hooks/pre.d/base-env2.yml",
            "env2/path/hooks/pre.d/env1-env2.yml",
            "env2/path/hooks/pre.d/env2-only.yml",
        ]
        actual = dispatcher._find_hooks(["config/path", "env1/path",
                                         "env2/path"], "pre")
        # Sort the result - it is not ordered at this stage.
        actual.sort()
        self.assertListEqual(actual, expected_result)

    @mock.patch.object(glob, 'glob')
    @mock.patch.object(os.path, 'exists')
    def test__find_hooks_non_existent(self, mock_exists, mock_glob):
        mock_exists.return_value = False
        mock_command = mock.MagicMock()
        dispatcher = commands.HookDispatcher(command=mock_command)
        expected_result = []
        actual = dispatcher._find_hooks(["config/path"], "pre")
        self.assertListEqual(actual, expected_result)
        mock_glob.assert_not_called()
